from module.models_onnx import SynthesizerTrn, symbols_v1, symbols_v2
from AR.models.t2s_lightning_module_onnx import Text2SemanticLightningModule
import torch
import torchaudio
from torch import nn
from feature_extractor import cnhubert

cnhubert_base_path = "moyoyo_tts/pretrained_models/chinese-hubert-base"
cnhubert.cnhubert_base_path = cnhubert_base_path
ssl_model = cnhubert.get_model()
from text import cleaned_text_to_sequence
import soundfile
import os
import json

def spectrogram_torch(y, n_fft, sampling_rate, hop_size, win_size, center=False):
    hann_window = torch.hann_window(win_size).to(
            dtype=y.dtype, device=y.device
        )
    y = torch.nn.functional.pad(
        y.unsqueeze(1),
        (int((n_fft - hop_size) / 2), int((n_fft - hop_size) / 2)),
        mode="reflect",
    )
    y = y.squeeze(1)
    spec = torch.stft(
        y,
        n_fft,
        hop_length=hop_size,
        win_length=win_size,
        window=hann_window,
        center=center,
        pad_mode="reflect",
        normalized=False,
        onesided=True,
        return_complex=False,
    )
    spec = torch.sqrt(spec.pow(2).sum(-1) + 1e-6)
    return spec


class DictToAttrRecursive(dict):
    def __init__(self, input_dict):
        super().__init__(input_dict)
        for key, value in input_dict.items():
            if isinstance(value, dict):
                value = DictToAttrRecursive(value)
            self[key] = value
            setattr(self, key, value)

    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            raise AttributeError(f"Attribute {item} not found")

    def __setattr__(self, key, value):
        if isinstance(value, dict):
            value = DictToAttrRecursive(value)
        super(DictToAttrRecursive, self).__setitem__(key, value)
        super().__setattr__(key, value)

    def __delattr__(self, item):
        try:
            del self[item]
        except KeyError:
            raise AttributeError(f"Attribute {item} not found")


class T2SEncoder(nn.Module):
    def __init__(self, t2s, vits):
        super().__init__()
        self.encoder = t2s.onnx_encoder
        self.vits = vits

    def forward(self, ref_seq, text_seq, ref_bert, text_bert, ssl_content):
        codes = self.vits.extract_latent(ssl_content)
        prompt_semantic = codes[0, 0]
        bert = torch.cat([ref_bert.transpose(0, 1), text_bert.transpose(0, 1)], 1)
        all_phoneme_ids = torch.cat([ref_seq, text_seq], 1)
        bert = bert.unsqueeze(0)
        prompt = prompt_semantic.unsqueeze(0)
        return self.encoder(all_phoneme_ids, bert), prompt


class T2SModel(nn.Module):
    def __init__(self, t2s_path, vits_model):
        super().__init__()
        dict_s1 = torch.load(t2s_path, map_location="cpu")
        self.config = dict_s1["config"]
        self.t2s_model = Text2SemanticLightningModule(self.config, "ojbk", is_train=False)
        self.t2s_model.load_state_dict(dict_s1["weight"])
        self.t2s_model.eval()
        self.vits_model = vits_model.vq_model
        self.hz = 50
        self.max_sec = self.config["data"]["max_sec"]
        self.t2s_model.model.top_k = torch.LongTensor([self.config["inference"]["top_k"]])
        self.t2s_model.model.early_stop_num = torch.LongTensor([self.hz * self.max_sec])
        self.t2s_model = self.t2s_model.model
        self.t2s_model.init_onnx()
        self.onnx_encoder = T2SEncoder(self.t2s_model, self.vits_model)
        self.first_stage_decoder = self.t2s_model.first_stage_decoder
        self.stage_decoder = self.t2s_model.stage_decoder
        #self.t2s_model = torch.jit.script(self.t2s_model)

    def forward(self, ref_seq, text_seq, ref_bert, text_bert, ssl_content):
        early_stop_num = self.t2s_model.early_stop_num

        #[1,N] [1,N] [N, 1024] [N, 1024] [1, 768, N]
        x, prompts = self.onnx_encoder(ref_seq, text_seq, ref_bert, text_bert, ssl_content)

        prefix_len = prompts.shape[1]

        #[1,N,512] [1,N]
        y, k, v, y_emb, x_example = self.first_stage_decoder(x, prompts)

        stop = False
        for idx in range(1, 1500):
            #[1, N] [N_layer, N, 1, 512] [N_layer, N, 1, 512] [1, N, 512] [1] [1, N, 512] [1, N]
            enco = self.stage_decoder(y, k, v, y_emb, x_example)
            y, k, v, y_emb, logits, samples = enco
            if early_stop_num != -1 and (y.shape[1] - prefix_len) > early_stop_num:
                stop = True
            if torch.argmax(logits, dim=-1)[0] == self.t2s_model.EOS or samples[0, 0] == self.t2s_model.EOS:
                stop = True
            if stop:
                break
        y[0, -1] = 0

        return y[:, -idx:].unsqueeze(0)

    def export(self, ref_seq, text_seq, ref_bert, text_bert, ssl_content, project_name, dynamo=False):
        #self.onnx_encoder = torch.jit.script(self.onnx_encoder)
        if dynamo:
            export_options = torch.onnx.ExportOptions(dynamic_shapes=True)
            onnx_encoder_export_output = torch.onnx.dynamo_export(
                self.onnx_encoder,
                (ref_seq, text_seq, ref_bert, text_bert, ssl_content),
                export_options=export_options
            )
            onnx_encoder_export_output.save(f"onnx/{project_name}/{project_name}_t2s_encoder.onnx")
            return

        torch.onnx.export(
            self.onnx_encoder,
            (ref_seq, text_seq, ref_bert, text_bert, ssl_content),
            f"onnx/{project_name}/{project_name}_t2s_encoder.onnx",
            input_names=["ref_seq", "text_seq", "ref_bert", "text_bert", "ssl_content"],
            output_names=["x", "prompts"],
            dynamic_axes={
                "ref_seq": {1 : "ref_length"},
                "text_seq": {1 : "text_length"},
                "ref_bert": {0 : "ref_length"},
                "text_bert": {0 : "text_length"},
                "ssl_content": {2 : "ssl_length"},
            },
            opset_version=16
        )
        x, prompts = self.onnx_encoder(ref_seq, text_seq, ref_bert, text_bert, ssl_content)

        torch.onnx.export(
            self.first_stage_decoder,
            (x, prompts),
            f"onnx/{project_name}/{project_name}_t2s_fsdec.onnx",
            input_names=["x", "prompts"],
            output_names=["y", "k", "v", "y_emb", "x_example"],
            dynamic_axes={
                "x": {1 : "x_length"},
                "prompts": {1 : "prompts_length"},
            },
            verbose=False,
            opset_version=16
        )
        y, k, v, y_emb, x_example = self.first_stage_decoder(x, prompts)

        torch.onnx.export(
            self.stage_decoder,
            (y, k, v, y_emb, x_example),
            f"onnx/{project_name}/{project_name}_t2s_sdec.onnx",
            input_names=["iy", "ik", "iv", "iy_emb", "ix_example"],
            output_names=["y", "k", "v", "y_emb", "logits", "samples"],
            dynamic_axes={
                "iy": {1 : "iy_length"},
                "ik": {1 : "ik_length"},
                "iv": {1 : "iv_length"},
                "iy_emb": {1 : "iy_emb_length"},
                "ix_example": {1 : "ix_example_length"},
            },
            verbose=False,
            opset_version=16
        )


class VitsModel(nn.Module):
    def __init__(self, vits_path):
        super().__init__()
        dict_s2 = torch.load(vits_path,map_location="cpu")
        self.hps = dict_s2["config"]
        if dict_s2['weight']['enc_p.text_embedding.weight'].shape[0] == 322:
            self.hps["model"]["version"] = "v1"
        else:
            self.hps["model"]["version"] = "v2"

        self.hps = DictToAttrRecursive(self.hps)
        self.hps.model.semantic_frame_rate = "25hz"
        self.vq_model = SynthesizerTrn(
            self.hps.data.filter_length // 2 + 1,
            self.hps.train.segment_size // self.hps.data.hop_length,
            n_speakers=self.hps.data.n_speakers,
            **self.hps.model
        )
        self.vq_model.eval()
        self.vq_model.load_state_dict(dict_s2["weight"], strict=False)

    def forward(self, text_seq, pred_semantic, ref_audio):
        refer = spectrogram_torch(
            ref_audio,
            self.hps.data.filter_length,
            self.hps.data.sampling_rate,
            self.hps.data.hop_length,
            self.hps.data.win_length,
            center=False
        )
        return self.vq_model(pred_semantic, text_seq, refer)[0, 0]


class GptSoVits(nn.Module):
    def __init__(self, vits, t2s):
        super().__init__()
        self.vits = vits
        self.t2s = t2s

    def forward(self, ref_seq, text_seq, ref_bert, text_bert, ref_audio, ssl_content, debug=False):
        pred_semantic = self.t2s(ref_seq, text_seq, ref_bert, text_bert, ssl_content)
        audio = self.vits(text_seq, pred_semantic, ref_audio)
        if debug:
            import onnxruntime
            sess = onnxruntime.InferenceSession("onnx/koharu/koharu_vits.onnx", providers=["CPU"])
            audio1 = sess.run(None, {
                "text_seq" : text_seq.detach().cpu().numpy(),
                "pred_semantic" : pred_semantic.detach().cpu().numpy(),
                "ref_audio" : ref_audio.detach().cpu().numpy()
            })
            return audio, audio1
        return audio

    def export(self, ref_seq, text_seq, ref_bert, text_bert, ref_audio, ssl_content, project_name):
        self.t2s.export(ref_seq, text_seq, ref_bert, text_bert, ssl_content, project_name)
        pred_semantic = self.t2s(ref_seq, text_seq, ref_bert, text_bert, ssl_content)
        torch.onnx.export(
            self.vits,
            (text_seq, pred_semantic, ref_audio),
            f"onnx/{project_name}/{project_name}_vits.onnx",
            input_names=["text_seq", "pred_semantic", "ref_audio"],
            output_names=["audio"],
            dynamic_axes={
                "text_seq": {1 : "text_length"},
                "pred_semantic": {2 : "pred_length"},
                "ref_audio": {1 : "audio_length"},
            },
            opset_version=17,
            verbose=False
        )


class SSLModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.ssl = ssl_model

    def forward(self, ref_audio_16k):
        return self.ssl.model(ref_audio_16k)["last_hidden_state"].transpose(1, 2)


def export(vits_path, gpt_path, project_name, vits_model="v2"):
    vits = VitsModel(vits_path)
    gpt = T2SModel(gpt_path, vits)
    gpt_sovits = GptSoVits(vits, gpt)
    ssl = SSLModel()
    ref_seq = torch.LongTensor([cleaned_text_to_sequence(["n", "i2", "h", "ao3", ",", "w", "o3", "sh", "i4", "b", "ai2", "y", "e4"],version=vits_model)])
    text_seq = torch.LongTensor([cleaned_text_to_sequence(["w", "o3", "sh", "i4", "b", "ai2", "y", "e4", "w", "o3", "sh", "i4", "b", "ai2", "y", "e4", "w", "o3", "sh", "i4", "b", "ai2", "y", "e4"],version=vits_model)])
    ref_bert = torch.randn((ref_seq.shape[1], 1024)).float()
    text_bert = torch.randn((text_seq.shape[1], 1024)).float()
    ref_audio = torch.randn((1, 48000 * 5)).float()
    # ref_audio = torch.tensor([load_audio("rec.wav", 48000)]).float()
    ref_audio_16k = torchaudio.functional.resample(ref_audio,48000,16000).float()
    ref_audio_sr = torchaudio.functional.resample(ref_audio,48000,vits.hps.data.sampling_rate).float()

    try:
        os.mkdir(f"onnx/{project_name}")
    except:
        pass

    ssl_content = ssl(ref_audio_16k).float()

    # debug = False
    debug = True

    # gpt_sovits.export(ref_seq, text_seq, ref_bert, text_bert, ref_audio_sr, ssl_content, project_name)

    if debug:
        a, b = gpt_sovits(ref_seq, text_seq, ref_bert, text_bert, ref_audio_sr, ssl_content, debug=debug)
        soundfile.write("out1.wav", a.cpu().detach().numpy(), vits.hps.data.sampling_rate)
        soundfile.write("out2.wav", b[0], vits.hps.data.sampling_rate)
    else:
        a = gpt_sovits(ref_seq, text_seq, ref_bert, text_bert, ref_audio_sr, ssl_content).detach().cpu().numpy()
        soundfile.write("out.wav", a, vits.hps.data.sampling_rate)

    if vits_model == "v1":
        symbols = symbols_v1
    else:
        symbols = symbols_v2

    MoeVSConf = {
        "Folder": f"{project_name}",
        "Name": f"{project_name}",
        "Type": "GPT-SoVits",
        "Rate": vits.hps.data.sampling_rate,
        "NumLayers": gpt.t2s_model.num_layers,
        "EmbeddingDim": gpt.t2s_model.embedding_dim,
        "Dict": "BasicDict",
        "BertPath": "chinese-roberta-wwm-ext-large",
        # "Symbol": symbols,
        "AddBlank": False,
    }

    MoeVSConfJson = json.dumps(MoeVSConf)
    with open(f"onnx/{project_name}.json", 'w') as MoeVsConfFile:
        json.dump(MoeVSConf, MoeVsConfFile, indent = 4)


if __name__ == "__main__":
    try:
        os.mkdir("onnx")
    except:
        pass

    gpt_path = "GPT_weights/nahida-e25.ckpt"
    vits_path = "SoVITS_weights/nahida_e30_s3930.pth"
    exp_path = "nahida"
    export(vits_path, gpt_path, exp_path)

    # soundfile.write("out.wav", a, vits.hps.data.sampling_rate)



# modified from https://github.com/yangdongchao/SoundStorm/blob/master/soundstorm/s1/AR/models/t2s_lightning_module.py
# reference: https://github.com/lifeiteng/vall-e
import os, sys

now_dir = os.getcwd()
sys.path.append(now_dir)
from typing import Dict

import torch
from pytorch_lightning import LightningModule
from moyoyo_tts.AR.models.t2s_model_onnx import Text2SemanticDecoder
from moyoyo_tts.AR.modules.lr_schedulers import WarmupCosineLRSchedule
from moyoyo_tts.AR.modules.optim import ScaledAdam


class Text2SemanticLightningModule(LightningModule):
    def __init__(self, config, output_dir, is_train=True):
        super().__init__()
        self.config = config
        self.top_k = 3
        self.model = Text2SemanticDecoder(config=config, top_k=self.top_k)
        pretrained_s1 = config.get("pretrained_s1")
        if pretrained_s1 and is_train:
            # print(self.load_state_dict(torch.load(pretrained_s1,map_location="cpu")["state_dict"]))
            print(
                self.load_state_dict(
                    torch.load(pretrained_s1, map_location="cpu")["weight"]
                )
            )
        if is_train:
            self.automatic_optimization = False
            self.save_hyperparameters()
            self.eval_dir = output_dir / "eval"
            self.eval_dir.mkdir(parents=True, exist_ok=True)

    def training_step(self, batch: Dict, batch_idx: int):
        opt = self.optimizers()
        scheduler = self.lr_schedulers()
        loss, acc = self.model.forward(
            batch["phoneme_ids"],
            batch["phoneme_ids_len"],
            batch["semantic_ids"],
            batch["semantic_ids_len"],
            batch["bert_feature"],
        )
        self.manual_backward(loss)
        if batch_idx > 0 and batch_idx % 4 == 0:
            opt.step()
            opt.zero_grad()
            scheduler.step()

        self.log(
            "total_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            sync_dist=True,
        )
        self.log(
            "lr",
            scheduler.get_last_lr()[0],
            on_epoch=True,
            prog_bar=True,
            sync_dist=True,
        )
        self.log(
            f"top_{self.top_k}_acc",
            acc,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            sync_dist=True,
        )

    def validation_step(self, batch: Dict, batch_idx: int):
        return

    def configure_optimizers(self):
        model_parameters = self.model.parameters()
        parameters_names = []
        parameters_names.append(
            [name_param_pair[0] for name_param_pair in self.model.named_parameters()]
        )
        lm_opt = ScaledAdam(
            model_parameters,
            lr=0.01,
            betas=(0.9, 0.95),
            clipping_scale=2.0,
            parameters_names=parameters_names,
            show_dominant_parameters=False,
            clipping_update_period=1000,
        )

        return {
            "optimizer": lm_opt,
            "lr_scheduler": {
                "scheduler": WarmupCosineLRSchedule(
                    lm_opt,
                    init_lr=self.config["optimizer"]["lr_init"],
                    peak_lr=self.config["optimizer"]["lr"],
                    end_lr=self.config["optimizer"]["lr_end"],
                    warmup_steps=self.config["optimizer"]["warmup_steps"],
                    total_steps=self.config["optimizer"]["decay_steps"],
                )
            },
        }



import os
from typing import Union


def init_proxy_env(is_proxy:bool=True, proxy_url:str= 'http://192.168.0.75:10809', ):
    if is_proxy:
        os.environ['http_proxy'] = proxy_url
        os.environ['https_proxy'] = proxy_url


def init_env(env:dict):
    for env_name, env_value in env.items():
        os.environ[env_name] = env_value


def set_api_keys(
    model_api_key: str,
    module_api_url: Union[str] = None,
    model_name: Union[str] = None,
    model_max_tokens: Union[int,str] = None
) -> None:
    os.environ["OPENAI_API_KEY"] = model_api_key
    if module_api_url:
        os.environ["OPENAI_API_BASE"] = module_api_url
    if model_name:
        os.environ["OPENAI_MODEL_NAME"] = model_name
    if model_max_tokens:
        os.environ["OPENAI_MAX_TOKENS"] = str(20480)



import os

from moyoyo_tts.text import symbols as symbols_v1
from moyoyo_tts.text import symbols2 as symbols_v2

# if os.environ.get("version","v1")=="v1":
#   from text.symbols import symbols
# else:
#   from text.symbols2 import symbols

_symbol_to_id_v1 = {s: i for i, s in enumerate(symbols_v1.symbols)}
_symbol_to_id_v2 = {s: i for i, s in enumerate(symbols_v2.symbols)}

def cleaned_text_to_sequence(cleaned_text, version=None):
  '''Converts a string of text to a sequence of IDs corresponding to the symbols in the text.
    Args:
      text: string to convert to a sequence
    Returns:
      List of integers corresponding to the symbols in the text
  '''
  if version is None:version=os.environ.get('version', 'v2')
  if version == "v1":
    phones = [_symbol_to_id_v1[symbol] for symbol in cleaned_text]
  else:
    phones = [_symbol_to_id_v2[symbol] for symbol in cleaned_text]

  return phones




"""
MoFA search command - Search for nodes and flows
"""

import os
import click
from mofa import agents_dir_path, flows_dir_path
from mofa.utils.files.dir import get_subdirectories
from mofa.registry import HubClient


def register_search_commands(cli_group):
    """Register search command group to the main CLI"""

    @cli_group.group(invoke_without_command=True)
    @click.pass_context
    def search(ctx):
        """Search for nodes and flows"""
        if ctx.invoked_subcommand is None:
            # No subcommand, run search TUI
            _run_search_tui()

    def _run_search_tui():
        """Run interactive search TUI"""
        click.echo("\n" + "=" * 50)
        click.echo("           MoFA Search")
        click.echo("=" * 50 + "\n")

        # Ask what to search
        search_type = click.prompt(
            "What to search? (1=agents, 2=flows, q=quit)", type=str, default="1"
        )

        if search_type.lower() == "q":
            return

        keyword = click.prompt("Search keyword", type=str)

        # Ask scope
        scope = click.prompt(
            "Search where? (1=local, 2=remote, 3=both)", type=str, default="3"
        )

        local_only = scope == "1"
        remote_only = scope == "2"

        click.echo()

        remote_results = []

        if search_type == "1":
            # Search agents
            local_agents = get_subdirectories(agents_dir_path)
            keyword_lower = keyword.lower()

            if not remote_only:
                matches = [
                    name for name in local_agents if keyword_lower in name.lower()
                ]
                if matches:
                    click.echo(f"Local agents matching '{keyword}' ({len(matches)}):")
                    for name in sorted(matches):
                        agent_path = os.path.join(agents_dir_path, name)
                        click.echo(f"  [local] {name}")
                        click.echo(f"         {agent_path}")

            if not local_only:
                try:
                    hub = HubClient()
                    remote_matches = hub.search_agents(keyword)
                    if remote_matches:
                        if not remote_only:
                            click.echo()
                        click.echo(
                            f"Remote agents matching '{keyword}' ({len(remote_matches)}):"
                        )
                        for idx, agent in enumerate(remote_matches, 1):
                            name = agent.get("name", "unknown")
                            desc = agent.get("description", "")
                            click.echo(f"  {idx}. [hub] {name}")
                            if desc:
                                click.echo(f"          {desc}")
                        remote_results = remote_matches

                        # Ask if user wants to download
                        if click.confirm(
                            "\nDownload any of these agents?", default=False
                        ):
                            choice = click.prompt("Select agent number", type=str)
                            try:
                                agent_idx = int(choice) - 1
                                if 0 <= agent_idx < len(remote_matches):
                                    selected_agent = remote_matches[agent_idx]["name"]
                                    output_dir = click.prompt(
                                        "Output directory", default=agents_dir_path
                                    )
                                    click.echo(f"\nDownloading '{selected_agent}'...")
                                    hub.download_agent(selected_agent, output_dir)
                                    click.echo(
                                        f"Successfully downloaded to {output_dir}/{selected_agent}"
                                    )
                                else:
                                    click.echo("Invalid selection")
                            except ValueError:
                                click.echo("Invalid input")
                            except Exception as e:
                                click.echo(f"Error: {e}", err=True)
                except Exception as e:
                    click.echo(f"Error searching remote: {e}", err=True)

        elif search_type == "2":
            # Search flows
            local_flows = get_subdirectories(flows_dir_path)
            keyword_lower = keyword.lower()

            if not remote_only:
                matches = [
                    name for name in local_flows if keyword_lower in name.lower()
                ]
                if matches:
                    click.echo(f"Local flows matching '{keyword}' ({len(matches)}):")
                    for name in sorted(matches):
                        flow_path = os.path.join(flows_dir_path, name)
                        click.echo(f"  [local] {name}")
                        click.echo(f"         {flow_path}")

            if not local_only:
                try:
                    hub = HubClient()
                    remote_matches = hub.search_flows(keyword)
                    if remote_matches:
                        if not remote_only:
                            click.echo()
                        click.echo(
                            f"Remote flows matching '{keyword}' ({len(remote_matches)}):"
                        )
                        for idx, flow in enumerate(remote_matches, 1):
                            name = flow.get("name", "unknown")
                            desc = flow.get("description", "")
                            click.echo(f"  {idx}. [hub] {name}")
                            if desc:
                                click.echo(f"          {desc}")
                        remote_results = remote_matches

                        # Ask if user wants to download
                        if click.confirm(
                            "\nDownload any of these flows?", default=False
                        ):
                            choice = click.prompt("Select flow number", type=str)
                            try:
                                flow_idx = int(choice) - 1
                                if 0 <= flow_idx < len(remote_matches):
                                    selected_flow = remote_matches[flow_idx]["name"]
                                    output_dir = click.prompt(
                                        "Output directory", default=flows_dir_path
                                    )
                                    click.echo(f"\nDownloading '{selected_flow}'...")
                                    hub.download_flow(selected_flow, output_dir)
                                    click.echo(
                                        f"Successfully downloaded to {output_dir}/{selected_flow}"
                                    )
                                else:
                                    click.echo("Invalid selection")
                            except ValueError:
                                click.echo("Invalid input")
                            except Exception as e:
                                click.echo(f"Error: {e}", err=True)
                except Exception as e:
                    click.echo(f"Error searching remote: {e}", err=True)

    @search.command()
    @click.argument("keyword", required=True)
    @click.option("--local", is_flag=True, help="Search only local agents")
    @click.option("--remote", is_flag=True, help="Search only remote hub agents")
    def agent(keyword, local, remote):
        """Search for agents (searches both local and remote by default)"""

        # Get local agents
        local_agents = get_subdirectories(agents_dir_path)
        keyword_lower = keyword.lower()

        if local:
            # Local only
            matches = [name for name in local_agents if keyword_lower in name.lower()]
            click.echo(f"Local agents matching '{keyword}' ({len(matches)}):")
            if matches:
                for name in sorted(matches):
                    agent_path = os.path.join(agents_dir_path, name)
                    click.echo(f"  - {name}")
                    click.echo(f"    {agent_path}")
            else:
                click.echo("  No matches found")
            return

        if remote:
            # Remote only
            try:
                hub = HubClient()
                matches = hub.search_agents(keyword)
                click.echo(f"Remote agents matching '{keyword}' ({len(matches)}):")
                if matches:
                    for agent in matches:
                        name = agent.get("name", "unknown")
                        desc = agent.get("description", "")
                        tags = ", ".join(agent.get("tags", []))
                        click.echo(f"  - {name}")
                        if desc:
                            click.echo(f"    {desc}")
                        if tags:
                            click.echo(f"    Tags: {tags}")
                else:
                    click.echo("  No matches found")
            except Exception as e:
                click.echo(f"Error searching remote agents: {e}", err=True)
            return

        # Both local and remote (default)
        local_matches = [name for name in local_agents if keyword_lower in name.lower()]

        if local_matches:
            click.echo(f"Local agents matching '{keyword}' ({len(local_matches)}):")
            for name in sorted(local_matches):
                agent_path = os.path.join(agents_dir_path, name)
                click.echo(f"  [local] {name}")
                click.echo(f"         {agent_path}")

        try:
            hub = HubClient()
            remote_matches = hub.search_agents(keyword)
            if remote_matches:
                if local_matches:
                    click.echo()
                click.echo(
                    f"Remote agents matching '{keyword}' ({len(remote_matches)}):"
                )
                for agent in remote_matches:
                    name = agent.get("name", "unknown")
                    desc = agent.get("description", "")
                    click.echo(f"  [hub] {name}")
                    if desc:
                        click.echo(f"       {desc}")
        except Exception as e:
            click.echo(f"\nError searching remote agents: {e}", err=True)

        if not local_matches:
            try:
                hub = HubClient()
                remote_matches = hub.search_agents(keyword)
                if not remote_matches:
                    click.echo(f"No agents found matching '{keyword}'")
            except:
                pass

    @search.command()
    @click.argument("keyword", required=True)
    @click.option("--local", is_flag=True, help="Search only local flows")
    @click.option("--remote", is_flag=True, help="Search only remote hub flows")
    def flow(keyword, local, remote):
        """Search for flows (searches both local and remote by default)"""

        # Get local flows
        local_flows = get_subdirectories(flows_dir_path)
        keyword_lower = keyword.lower()

        if local:
            # Local only
            matches = [name for name in local_flows if keyword_lower in name.lower()]
            click.echo(f"Local flows matching '{keyword}' ({len(matches)}):")
            if matches:
                for name in sorted(matches):
                    flow_path = os.path.join(flows_dir_path, name)
                    click.echo(f"  - {name}")
                    click.echo(f"    {flow_path}")
            else:
                click.echo("  No matches found")
            return

        if remote:
            # Remote only
            try:
                hub = HubClient()
                matches = hub.search_flows(keyword)
                click.echo(f"Remote flows matching '{keyword}' ({len(matches)}):")
                if matches:
                    for flow in matches:
                        name = flow.get("name", "unknown")
                        desc = flow.get("description", "")
                        agents = ", ".join(flow.get("agents", []))
                        click.echo(f"  - {name}")
                        if desc:
                            click.echo(f"    {desc}")
                        if agents:
                            click.echo(f"    Agents: {agents}")
                else:
                    click.echo("  No matches found")
            except Exception as e:
                click.echo(f"Error searching remote flows: {e}", err=True)
            return

        # Both local and remote (default)
        local_matches = [name for name in local_flows if keyword_lower in name.lower()]

        if local_matches:
            click.echo(f"Local flows matching '{keyword}' ({len(local_matches)}):")
            for name in sorted(local_matches):
                flow_path = os.path.join(flows_dir_path, name)
                click.echo(f"  [local] {name}")
                click.echo(f"         {flow_path}")

        try:
            hub = HubClient()
            remote_matches = hub.search_flows(keyword)
            if remote_matches:
                if local_matches:
                    click.echo()
                click.echo(
                    f"Remote flows matching '{keyword}' ({len(remote_matches)}):"
                )
                for flow in remote_matches:
                    name = flow.get("name", "unknown")
                    desc = flow.get("description", "")
                    click.echo(f"  [hub] {name}")
                    if desc:
                        click.echo(f"       {desc}")
        except Exception as e:
            click.echo(f"\nError searching remote flows: {e}", err=True)

        if not local_matches:
            try:
                hub = HubClient()
                remote_matches = hub.search_flows(keyword)
                if not remote_matches:
                    click.echo(f"No flows found matching '{keyword}'")
            except:
                pass



import sys

import torch
import yaml


def load_yaml_config(path):
    with open(path) as f:
        config = yaml.full_load(f)
    return config


def save_config_to_yaml(config, path):
    assert path.endswith(".yaml")
    with open(path, "w") as f:
        f.write(yaml.dump(config))
        f.close()


def write_args(args, path):
    args_dict = dict(
        (name, getattr(args, name)) for name in dir(args) if not name.startswith("_")
    )
    with open(path, "a") as args_file:
        args_file.write("==> torch version: {}\n".format(torch.__version__))
        args_file.write(
            "==> cudnn version: {}\n".format(torch.backends.cudnn.version())
        )
        args_file.write("==> Cmd:\n")
        args_file.write(str(sys.argv))
        args_file.write("\n==> args:\n")
        for k, v in sorted(args_dict.items()):
            args_file.write("  %s: %s\n" % (str(k), str(v)))
        args_file.close()



"""
Configuration for PrimeSpeech TTS node.
"""

import os
from pathlib import Path
from typing import Optional, Dict, Any
from dataclasses import dataclass, field


@dataclass
class PrimeSpeechConfig:
    """PrimeSpeech configuration class"""
    
    # Model selection
    VOICE_NAME = os.getenv("VOICE_NAME", "Doubao")  # Voice character name
    REPOSITORY = os.getenv("REPOSITORY", "MoYoYoTech/tone-models")  # HuggingFace repo
    
    # Model paths
    MODEL_DIR = os.getenv("PRIMESPEECH_MODEL_DIR", "")  # Override model directory
    GPT_MODEL_PATH = os.getenv("GPT_MODEL_PATH", "")  # Override GPT model path
    SOVITS_MODEL_PATH = os.getenv("SOVITS_MODEL_PATH", "")  # Override SoVITS model path
    REFERENCE_AUDIO_PATH = os.getenv("REFERENCE_AUDIO_PATH", "")  # Override reference audio
    
    # Language settings
    TEXT_LANG = os.getenv("TEXT_LANG", "auto")  # auto, zh, en, ja
    PROMPT_LANG = os.getenv("PROMPT_LANG", "auto")  # auto, zh, en, ja
    PROMPT_TEXT = os.getenv("PROMPT_TEXT", "")  # Optional custom prompt text
    
    # Inference parameters (optimized for streaming)
    TOP_K = int(os.getenv("TOP_K", "3"))  # Lower for faster inference
    TOP_P = float(os.getenv("TOP_P", "0.95"))
    TEMPERATURE = float(os.getenv("TEMPERATURE", "0.8"))  # Lower for consistency
    SPEED_FACTOR_OVERRIDE: Optional[str] = os.getenv("SPEED_FACTOR")
    SPEED_FACTOR: Optional[float] = (
        float(SPEED_FACTOR_OVERRIDE) if SPEED_FACTOR_OVERRIDE is not None else None
    )
    BATCH_SIZE = int(os.getenv("BATCH_SIZE", "100"))
    SEED = int(os.getenv("SEED", "233333"))
    
    # Processing options
    TEXT_SPLIT_METHOD = os.getenv("TEXT_SPLIT_METHOD", "cut5")  # cut0, cut1, cut2, cut3, cut4, cut5
    SPLIT_BUCKET = os.getenv("SPLIT_BUCKET", "true").lower() == "true"
    RETURN_FRAGMENT = os.getenv("RETURN_FRAGMENT", "false").lower() == "true"  # Respect env variable
    FRAGMENT_INTERVAL = float(os.getenv("FRAGMENT_INTERVAL", "0.3"))
    
    # Performance
    USE_GPU = os.getenv("USE_GPU", "false").lower() == "true"
    DEVICE = os.getenv("DEVICE", "cuda" if USE_GPU else "cpu")
    NUM_THREADS = int(os.getenv("NUM_THREADS", "4"))
    
    # Audio settings
    SAMPLE_RATE = int(os.getenv("SAMPLE_RATE", "32000"))
    
    # Logging
    LOG_LEVEL = os.getenv("LOG_LEVEL", "INFO").upper()  # DEBUG, INFO, WARNING, ERROR
    
    @staticmethod
    def get_models_dir() -> Path:
        """Get models directory path"""
        # Check environment variable first
        models_dir = os.getenv("PRIMESPEECH_MODEL_DIR")
        if models_dir:
            # Expand env vars (e.g., $HOME) and user (~)
            expanded = os.path.expanduser(os.path.expandvars(models_dir))
            return Path(expanded)
        
        # Default to home directory
        home = Path.home()
        default_dir = home / ".dora" / "models" / "primespeech"
        default_dir.mkdir(parents=True, exist_ok=True)
        return default_dir


# Voice configurations with default prompts and settings
VOICE_CONFIGS = {
    "Doubao": {
        "repository": "MoYoYoTech/tone-models",
        "gpt_weights": "GPT_weights/doubao_best_gpt.ckpt",
        "sovits_weights": "SoVITS_weights/doubao_best_sovits.pth",
        "reference_audio": "ref_audios/doubao_ref_mix_new.wav",
        "prompt_text": "这家resturant的steak很有名，但是vegetable salad的price有点贵",
        "text_lang": "zh",
        "prompt_lang": "zh",
        "speed_factor": 1.1,
    },
    "Luo Xiang": {
        "repository": "MoYoYoTech/tone-models",
        "gpt_weights": "GPT_weights/luoxiang_best_gpt.ckpt",
        "sovits_weights": "SoVITS_weights/luoxiang_best_sovits.pth",
        "reference_audio": "ref_audios/luoxiang_ref.wav",
        "prompt_text": "复杂的问题背后也许没有统一的答案，选择站在正方还是反方，其实取决于你对一系列价值判断的回答。",
        "text_lang": "zh",
        "prompt_lang": "zh",
        "speed_factor": 1.1,
    },
    "Yang Mi": {
        "repository": "MoYoYoTech/tone-models",
        "gpt_weights": "GPT_weights/yangmi_best_gpt.ckpt",
        "sovits_weights": "SoVITS_weights/yangmi_best_sovits.pth",
        "reference_audio": "ref_audios/yangmi_ref.wav",
        "prompt_text": "你谁知道, 人生只有一次啊. 你怎么知道那样选, 你当下来说, 应该那样选.",
        "text_lang": "zh",
        "prompt_lang": "zh",
        "speed_factor": 1.1,
    },
    "Zhou Jielun": {
        "repository": "MoYoYoTech/tone-models",
        "gpt_weights": "GPT_weights/zhoujielun_best_gpt.ckpt",
        "sovits_weights": "SoVITS_weights/zhoujielun_best_sovits.pth",
        "reference_audio": "ref_audios/zhoujielun_ref.wav",
        "prompt_text": "其实我我现在讲的这些奥，都是我未来成功的一些关键。",
        "text_lang": "zh",
        "prompt_lang": "zh",
        "speed_factor": 1.1,
    },
    "Ma Yun": {
        "repository": "MoYoYoTech/tone-models",
        "gpt_weights": "GPT_weights/mayun_best_gpt.ckpt",
        "sovits_weights": "SoVITS_weights/mayun_best_sovits.pth",
        "reference_audio": "ref_audios/mayun_ref.wav",
        "prompt_text": "这是我们最大的希望能招聘的到人。所以今天阿里巴巴公司内部，我自己这么觉得，人才梯队的建设非常之好。",
        "text_lang": "zh",
        "prompt_lang": "zh",
        "speed_factor": 1.1,
    },
    "Chen Yifan": {
        "repository": "MoYoYoTech/tone-models",
        "gpt_weights": "GPT_weights/yfc_best_gpt.ckpt",
        "sovits_weights": "SoVITS_weights/yfc_best_sovits.pth",
        "reference_audio": "ref_audios/yfc_ref.wav",
        "prompt_text": "他们的一个专家Marcel Marini的观点，他就提醒说这波投资热潮可能有点哎，我们说过热的迹象。",
        "text_lang": "zh",
        "prompt_lang": "zh",
        "speed_factor": 1.1,
    },
    "Zhao Daniu": {
        "repository": "MoYoYoTech/tone-models",
        "gpt_weights": "GPT_weights/dnz_best_gpt.ckpt",
        "sovits_weights": "SoVITS_weights/dnz_best_sovits.pth",
        "reference_audio": "ref_audios/dnz_ref.wav",
        "prompt_text": "今天啊我们要跟你一起深入探讨一篇嗯，来自经济学院的the economist的文章。",
        "text_lang": "zh",
        "prompt_lang": "zh",
        "speed_factor": 1.1,
    },
    "Maple": {
        "repository": "MoYoYoTech/tone-models",
        "gpt_weights": "GPT_weights/maple_best_gpt.ckpt",
        "sovits_weights": "SoVITS_weights/maple_best_sovits.pth",
        "reference_audio": "ref_audios/maple_ref.wav",
        "prompt_text": "There was a little tea shop in a bustling village. Every morning, the owner, an elderly woman, would wake up early.",
        "text_lang": "en",
        "prompt_lang": "en",
        "speed_factor": 1.0,
    },
    "Cove": {
        "repository": "MoYoYoTech/tone-models",
        "gpt_weights": "GPT_weights/cove_best_gpt.ckpt",
        "sovits_weights": "SoVITS_weights/cove_best_sovits.pth",
        "reference_audio": "ref_audios/cove_ref.wav",
        "prompt_text": "and he has a long career in the Senate representing Delaware. So both have had significant impacts on American politics and policies.",
        "text_lang": "en",
        "prompt_lang": "en",
        "speed_factor": 1.0,
    },
    "BYS": {
        "repository": "MoYoYoTech/tone-models",
        "gpt_weights": "GPT_weights/bys_best_gpt.ckpt",
        "sovits_weights": "SoVITS_weights/bys_best_sovits.pth",
        "reference_audio": "ref_audios/bys_ref.wav",
        "prompt_text": "今天天气不错，适合出去走走。",
        "text_lang": "zh",
        "prompt_lang": "zh",
        "speed_factor": 1.1,
    },
    "Ellen": {
        "repository": "MoYoYoTech/tone-models",
        "gpt_weights": "GPT_weights/ellen_best_gpt.ckpt",
        "sovits_weights": "SoVITS_weights/ellen_best_sovits.pth",
        "reference_audio": "ref_audios/ellen_ref.wav",
        "prompt_text": "Welcome to the show! Today we have some amazing guests.",
        "text_lang": "en",
        "prompt_lang": "en",
        "speed_factor": 1.1,
    },
    "Juniper": {
        "repository": "MoYoYoTech/tone-models",
        "gpt_weights": "GPT_weights/juniper_best_gpt.ckpt",
        "sovits_weights": "SoVITS_weights/juniper_best_sovits.pth",
        "reference_audio": "ref_audios/juniper_ref.wav",
        "prompt_text": "The forest was quiet, with only the sound of leaves rustling in the wind.",
        "text_lang": "en",
        "prompt_lang": "en",
        "speed_factor": 1.1,
    },
    "Ma Baoguo": {
        "repository": "MoYoYoTech/tone-models",
        "gpt_weights": "GPT_weights/mabaoguo_best_gpt.ckpt",
        "sovits_weights": "SoVITS_weights/mabaoguo_best_sovits.pth",
        "reference_audio": "ref_audios/mabaoguo_ref.wav",
        "prompt_text": "年轻人不讲武德，偷袭我这个六十九岁的老同志。",
        "text_lang": "zh",
        "prompt_lang": "zh",
        "speed_factor": 1.1,
    },
    "Shen Yi": {
        "repository": "MoYoYoTech/tone-models",
        "gpt_weights": "GPT_weights/shenyi_best_gpt.ckpt",
        "sovits_weights": "SoVITS_weights/shenyi_best_sovits.pth",
        "reference_audio": "ref_audios/shenyi_ref.wav",
        "prompt_text": "今天我们来分析一下这个案例的关键点。",
        "text_lang": "zh",
        "prompt_lang": "zh",
        "speed_factor": 1.1,
    },
    "Trump": {
        "repository": "MoYoYoTech/tone-models",
        "gpt_weights": "GPT_weights/trump_best_gpt.ckpt",
        "sovits_weights": "SoVITS_weights/trump_best_sovits.pth",
        "reference_audio": "ref_audios/trump_ref.wav",
        "prompt_text": "We're going to make America great again, and it's going to be tremendous.",
        "text_lang": "en",
        "prompt_lang": "en",
        "speed_factor": 1.1,
    },
}



# modified from https://github.com/CjangCjengh/vits/blob/main/text/japanese.py
import re
import os
import hashlib
try:
    import pyopenjtalk
    current_file_path = os.path.dirname(__file__)
    def get_hash(fp: str) -> str:
        hash_md5 = hashlib.md5()
        with open(fp, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    USERDIC_CSV_PATH = os.path.join(current_file_path, "ja_userdic", "userdict.csv")
    USERDIC_BIN_PATH = os.path.join(current_file_path, "ja_userdic", "user.dict")
    USERDIC_HASH_PATH = os.path.join(current_file_path, "ja_userdic", "userdict.md5")
    # 如果没有用户词典，就生成一个；如果有，就检查md5，如果不一样，就重新生成
    if os.path.exists(USERDIC_CSV_PATH):
        if not os.path.exists(USERDIC_BIN_PATH) or get_hash(USERDIC_CSV_PATH) != open(USERDIC_HASH_PATH, "r",encoding='utf-8').read():
            pyopenjtalk.mecab_dict_index(USERDIC_CSV_PATH, USERDIC_BIN_PATH)
            with open(USERDIC_HASH_PATH, "w", encoding='utf-8') as f:
                f.write(get_hash(USERDIC_CSV_PATH))

    if os.path.exists(USERDIC_BIN_PATH):
        pyopenjtalk.update_global_jtalk_with_user_dict(USERDIC_BIN_PATH)
except Exception as e:
    # print(e)
    import pyopenjtalk
    # failed to load user dictionary, ignore.
    pass


from moyoyo_tts.text.symbols import punctuation
# Regular expression matching Japanese without punctuation marks:
_japanese_characters = re.compile(
    r"[A-Za-z\d\u3005\u3040-\u30ff\u4e00-\u9fff\uff11-\uff19\uff21-\uff3a\uff41-\uff5a\uff66-\uff9d]"
)

# Regular expression matching non-Japanese characters or punctuation marks:
_japanese_marks = re.compile(
    r"[^A-Za-z\d\u3005\u3040-\u30ff\u4e00-\u9fff\uff11-\uff19\uff21-\uff3a\uff41-\uff5a\uff66-\uff9d]"
)

# List of (symbol, Japanese) pairs for marks:
_symbols_to_japanese = [(re.compile("%s" % x[0]), x[1]) for x in [("％", "パーセント")]]


# List of (consonant, sokuon) pairs:
_real_sokuon = [
    (re.compile("%s" % x[0]), x[1])
    for x in [
        (r"Q([↑↓]*[kg])", r"k#\1"),
        (r"Q([↑↓]*[tdjʧ])", r"t#\1"),
        (r"Q([↑↓]*[sʃ])", r"s\1"),
        (r"Q([↑↓]*[pb])", r"p#\1"),
    ]
]

# List of (consonant, hatsuon) pairs:
_real_hatsuon = [
    (re.compile("%s" % x[0]), x[1])
    for x in [
        (r"N([↑↓]*[pbm])", r"m\1"),
        (r"N([↑↓]*[ʧʥj])", r"n^\1"),
        (r"N([↑↓]*[tdn])", r"n\1"),
        (r"N([↑↓]*[kg])", r"ŋ\1"),
    ]
]


def post_replace_ph(ph):
    rep_map = {
        "：": ",",
        "；": ",",
        "，": ",",
        "。": ".",
        "！": "!",
        "？": "?",
        "\n": ".",
        "·": ",",
        "、": ",",
        "...": "…",
    }

    if ph in rep_map.keys():
        ph = rep_map[ph]
    return ph


def replace_consecutive_punctuation(text):
    punctuations = ''.join(re.escape(p) for p in punctuation)
    pattern = f'([{punctuations}])([{punctuations}])+'
    result = re.sub(pattern, r'\1', text)
    return result


def symbols_to_japanese(text):
    for regex, replacement in _symbols_to_japanese:
        text = re.sub(regex, replacement, text)
    return text


def preprocess_jap(text, with_prosody=False):
    """Reference https://r9y9.github.io/ttslearn/latest/notebooks/ch10_Recipe-Tacotron.html"""
    text = symbols_to_japanese(text)
    # English words to lower case, should have no influence on japanese words.
    text = text.lower()
    sentences = re.split(_japanese_marks, text)
    marks = re.findall(_japanese_marks, text)
    text = []
    for i, sentence in enumerate(sentences):
        if re.match(_japanese_characters, sentence):
            if with_prosody:
                text += pyopenjtalk_g2p_prosody(sentence)[1:-1]
            else:
                p = pyopenjtalk.g2p(sentence)
                text += p.split(" ")

        if i < len(marks):
            if marks[i] == " ":# 防止意外的UNK
                continue
            text += [marks[i].replace(" ", "")]
    return text


def text_normalize(text):
    # todo: jap text normalize

    # 避免重复标点引起的参考泄露
    text = replace_consecutive_punctuation(text)
    return text

# Copied from espnet https://github.com/espnet/espnet/blob/master/espnet2/text/phoneme_tokenizer.py
def pyopenjtalk_g2p_prosody(text, drop_unvoiced_vowels=True):
    """Extract phoneme + prosoody symbol sequence from input full-context labels.

    The algorithm is based on `Prosodic features control by symbols as input of
    sequence-to-sequence acoustic modeling for neural TTS`_ with some r9y9's tweaks.

    Args:
        text (str): Input text.
        drop_unvoiced_vowels (bool): whether to drop unvoiced vowels.

    Returns:
        List[str]: List of phoneme + prosody symbols.

    Examples:
        >>> from espnet2.text.phoneme_tokenizer import pyopenjtalk_g2p_prosody
        >>> pyopenjtalk_g2p_prosody("こんにちは。")
        ['^', 'k', 'o', '[', 'N', 'n', 'i', 'ch', 'i', 'w', 'a', '$']

    .. _`Prosodic features control by symbols as input of sequence-to-sequence acoustic
        modeling for neural TTS`: https://doi.org/10.1587/transinf.2020EDP7104

    """
    labels = pyopenjtalk.make_label(pyopenjtalk.run_frontend(text))
    N = len(labels)

    phones = []
    for n in range(N):
        lab_curr = labels[n]

        # current phoneme
        p3 = re.search(r"\-(.*?)\+", lab_curr).group(1)
        # deal unvoiced vowels as normal vowels
        if drop_unvoiced_vowels and p3 in "AEIOU":
            p3 = p3.lower()

        # deal with sil at the beginning and the end of text
        if p3 == "sil":
            assert n == 0 or n == N - 1
            if n == 0:
                phones.append("^")
            elif n == N - 1:
                # check question form or not
                e3 = _numeric_feature_by_regex(r"!(\d+)_", lab_curr)
                if e3 == 0:
                    phones.append("$")
                elif e3 == 1:
                    phones.append("?")
            continue
        elif p3 == "pau":
            phones.append("_")
            continue
        else:
            phones.append(p3)

        # accent type and position info (forward or backward)
        a1 = _numeric_feature_by_regex(r"/A:([0-9\-]+)\+", lab_curr)
        a2 = _numeric_feature_by_regex(r"\+(\d+)\+", lab_curr)
        a3 = _numeric_feature_by_regex(r"\+(\d+)/", lab_curr)

        # number of mora in accent phrase
        f1 = _numeric_feature_by_regex(r"/F:(\d+)_", lab_curr)

        a2_next = _numeric_feature_by_regex(r"\+(\d+)\+", labels[n + 1])
        # accent phrase border
        if a3 == 1 and a2_next == 1 and p3 in "aeiouAEIOUNcl":
            phones.append("#")
        # pitch falling
        elif a1 == 0 and a2_next == a2 + 1 and a2 != f1:
            phones.append("]")
        # pitch rising
        elif a2 == 1 and a2_next == 2:
            phones.append("[")

    return phones

# Copied from espnet https://github.com/espnet/espnet/blob/master/espnet2/text/phoneme_tokenizer.py
def _numeric_feature_by_regex(regex, s):
    match = re.search(regex, s)
    if match is None:
        return -50
    return int(match.group(1))

def g2p(norm_text, with_prosody=True):
    phones = preprocess_jap(norm_text, with_prosody)
    phones = [post_replace_ph(i) for i in phones]
    # todo: implement tones and word2ph
    return phones


if __name__ == "__main__":
    phones = g2p("Hello.こんにちは！今日もNiCe天気ですね！tokyotowerに行きましょう！")
    print(phones)



"""Standalone flow generator - reads agents and generates flows"""

import os
from pathlib import Path
from typing import Dict, List, Optional
import re


class FlowGenerator:
    """Generate MoFA flows from available agents"""

    def __init__(self, agents_dir: str, flows_dir: str, llm_model: str = "gpt-4o-mini", api_key: Optional[str] = None):
        """
        Initialize flow generator

        Args:
            agents_dir: Path to agents directory
            flows_dir: Path to flows output directory
            llm_model: LLM model to use
            api_key: API key (if None, reads from env)
        """
        self.agents_dir = Path(agents_dir)
        self.flows_dir = Path(flows_dir)
        self.llm_model = llm_model
        self.api_key = api_key or os.getenv('OPENAI_API_KEY')

        if not self.api_key:
            raise ValueError("API key not found. Set OPENAI_API_KEY or pass api_key parameter.")

    def scan_agents(self) -> Dict[str, Dict]:
        """
        Scan agents directory and extract information about each agent

        Returns:
            Dict mapping agent_name -> {
                'path': str,
                'code': str,
                'inputs': List[str],
                'outputs': List[str],
                'description': str
            }
        """
        agents_info = {}

        if not self.agents_dir.exists():
            return agents_info

        for agent_path in self.agents_dir.iterdir():
            if not agent_path.is_dir() or agent_path.name.startswith('.'):
                continue

            # Find main.py (recursively, max depth 2)
            main_py_path = None
            try:
                for f in agent_path.rglob("main.py"):
                    if len(f.relative_to(agent_path).parts) <= 2:
                        main_py_path = f
                        break
            except:
                continue

            if not main_py_path:
                continue

            # Read agent code
            try:
                with open(main_py_path, 'r', encoding='utf-8') as f:
                    code = f.read()

                # Extract metadata
                info = self._extract_agent_info(code, str(agent_path))
                agents_info[agent_path.name] = info

            except Exception:
                # Skip agents that can't be read
                continue

        return agents_info

    def _extract_agent_info(self, code: str, agent_path: str) -> Dict:
        """Extract information from agent code"""

        # Extract input parameters
        inputs = re.findall(r'receive_parameter\([\'"](\w+)[\'"]\)', code)
        params_match = re.findall(r'receive_parameters\(\[(.*?)\]\)', code, re.DOTALL)
        for match in params_match:
            param_names = re.findall(r'[\'"](\w+)[\'"]', match)
            inputs.extend(param_names)
        inputs = list(set(inputs))

        # Extract output parameters
        outputs = re.findall(r'send_output\(agent_output_name=[\'"](\w+)[\'"]', code)
        outputs = list(set(outputs))

        # Extract docstring
        docstring_match = re.search(r'"""(.*?)"""', code, re.DOTALL)
        description = ""
        if docstring_match:
            desc = docstring_match.group(1).strip()
            # Take first line or first 200 chars
            first_line = desc.split('\n')[0].strip()
            description = first_line[:200] if len(first_line) > 0 else desc[:200]

        return {
            'path': agent_path,
            'code': code,
            'inputs': inputs,
            'outputs': outputs,
            'description': description
        }

    def generate_flow(self, requirement: str, flow_name: Optional[str] = None, interactive: bool = True) -> str:
        """
        Generate a flow from requirement

        Args:
            requirement: What the flow should do
            flow_name: Optional flow name (generated if not provided)
            interactive: Enable interactive refinement (default: True)

        Returns:
            Path to generated flow directory
        """
        from .llm_client import LLMClient
        from .flow_scaffolder import FlowScaffolder

        # Scan agents
        agents_info = self.scan_agents()

        if not agents_info:
            raise RuntimeError(f"No agents found in {self.agents_dir}. Create agents first.")

        # Initialize LLM client
        client = LLMClient(model=self.llm_model, api_key=self.api_key)

        # Generate flow name if not provided
        if not flow_name:
            flow_name = client.generate_agent_name(requirement)

        # Build agents context for LLM
        agents_context = self._build_agents_context(agents_info)

        # Generate flow YAML
        flow_yaml = self._generate_flow_yaml(client, requirement, flow_name, agents_context)

        # Interactive refinement if enabled
        if interactive:
            flow_yaml = self._interactive_refinement(client, flow_yaml, requirement, agents_context)

        # Extract agent names from YAML
        agent_names = self._extract_agent_names(flow_yaml)

        # Create flow project
        scaffolder = FlowScaffolder(output_dir=str(self.flows_dir))
        flow_path = scaffolder.create_flow(flow_name, flow_yaml, agent_names)

        return flow_path

    def _interactive_refinement(self, client, flow_yaml: str, requirement: str, agents_context: str) -> str:
        """
        Allow user to interactively refine the flow through conversation

        Args:
            client: LLM client
            flow_yaml: Initial flow YAML
            requirement: Original requirement
            agents_context: Context about available agents

        Returns:
            Final refined flow YAML
        """
        import yaml

        while True:
            # Show current flow
            print("\n" + "=" * 60)
            print("Generated Flow YAML:")
            print("=" * 60)
            print(flow_yaml)
            print("=" * 60)

            # Validate YAML
            try:
                yaml.safe_load(flow_yaml)
                print("\n✓ Valid YAML")
            except Exception as e:
                print(f"\n⚠ YAML validation failed: {e}")

            # Ask what to do
            print("\nOptions:")
            print("  1. Accept and create flow")
            print("  2. Modify with AI (describe changes)")
            print("  3. Regenerate from scratch")
            print("  q. Cancel")

            choice = input("\nChoice [1]: ").strip() or "1"

            if choice == "1":
                # Accept
                return flow_yaml
            elif choice.lower() == "q":
                raise KeyboardInterrupt("User cancelled")
            elif choice == "3":
                # Regenerate
                print("\nRegenerating flow...")
                flow_yaml = self._generate_flow_yaml(client, requirement, "", agents_context)
            elif choice == "2":
                # Conversational modification
                print("\n" + "=" * 60)
                print("Conversational Flow Editor")
                print("=" * 60)
                print("Describe the changes you want. Type 'done' when finished.\n")

                conversation_history = []

                while True:
                    try:
                        user_input = input("You: ").strip()
                    except EOFError:
                        break

                    if not user_input:
                        continue

                    if user_input.lower() in ['done', 'finish', 'ok']:
                        break

                    # Modify flow
                    print("AI is thinking...")
                    try:
                        modified_yaml = self._modify_flow_conversational(
                            client, flow_yaml, user_input, agents_context, conversation_history
                        )

                        # Validate
                        try:
                            yaml.safe_load(modified_yaml)
                            flow_yaml = modified_yaml
                            print("\n✓ Flow updated")
                        except Exception as e:
                            print(f"\n⚠ Modified YAML is invalid: {e}")
                            print("Keeping previous version.")

                        # Show updated flow
                        print("\nUpdated flow:")
                        print("-" * 60)
                        print(flow_yaml)
                        print("-" * 60)

                        # Update conversation history
                        conversation_history.append({"user": user_input, "yaml": flow_yaml})

                    except Exception as e:
                        print(f"\n✗ Error: {e}")

                # After conversation, show final version and loop back
                continue
            else:
                print("Invalid choice. Please try again.")

        return flow_yaml

    def _modify_flow_conversational(self, client, current_yaml: str, user_instruction: str,
                                    agents_context: str, conversation_history: list) -> str:
        """Modify flow YAML based on user instruction"""

        history_str = ""
        if conversation_history:
            history_str = "\nPrevious modifications:\n"
            for i, item in enumerate(conversation_history[-3:], 1):  # Last 3 changes
                history_str += f"{i}. User: {item['user']}\n"

        prompt = f"""Modify the following MoFA flow YAML according to the user's instruction.

Current YAML:
```yaml
{current_yaml}
```
{history_str}

{agents_context}

User instruction: {user_instruction}

IMPORTANT Rules:
1. Only use agents from the "Available Agents" section above
2. Maintain valid YAML structure
3. Keep the same flow format (nodes with id, build, path, inputs, outputs, env)
4. Ensure proper input/output connections
5. Last node must have IS_DATAFLOW_END: true in env

Output ONLY the complete modified YAML, no explanations or markdown:
"""

        yaml_str = client.generate(prompt)

        # Clean up markdown
        yaml_str = re.sub(r'```yaml\n', '', yaml_str)
        yaml_str = re.sub(r'```\n?', '', yaml_str)

        return yaml_str.strip()

    def _build_agents_context(self, agents_info: Dict) -> str:
        """Build context string with all agents info for LLM"""

        context_parts = ["Available agents:\n"]

        for agent_name, info in sorted(agents_info.items()):
            context_parts.append(f"\n## Agent: {agent_name}")
            context_parts.append(f"Path: {info['path']}")

            if info['description']:
                context_parts.append(f"Description: {info['description']}")

            if info['inputs']:
                context_parts.append(f"Inputs: {', '.join(info['inputs'])}")
            else:
                context_parts.append("Inputs: (none)")

            if info['outputs']:
                context_parts.append(f"Outputs: {', '.join(info['outputs'])}")
            else:
                context_parts.append("Outputs: (none)")

            # Include code snippet (first 50 lines or 2000 chars)
            code_lines = info['code'].split('\n')[:50]
            code_snippet = '\n'.join(code_lines)[:2000]
            context_parts.append(f"\nCode:\n```python\n{code_snippet}\n```")

        return '\n'.join(context_parts)

    def _generate_flow_yaml(self, client, requirement: str, flow_name: str, agents_context: str) -> str:
        """Generate flow YAML using LLM"""

        prompt = f"""Generate a MoFA flow dataflow YAML based on this requirement:

## Requirement
{requirement}

## Flow Name
{flow_name}

## Available Agents
{agents_context}

## Flow YAML Structure
A flow is a YAML file that defines nodes (agents) and their connections.

Example:
```yaml
nodes:
  - id: terminal-input
    build: pip install -e ../../agents/terminal-input
    path: dynamic
    outputs:
      - data
    inputs:
      result: agent2/final_output

  - id: agent1
    build: pip install -e ../../agents/agent1
    path: agent1
    outputs:
      - processed_data
    inputs:
      raw_input: terminal-input/data

  - id: agent2
    build: pip install -e ../../agents/agent2
    path: agent2
    outputs:
      - final_output
    inputs:
      data: agent1/processed_data
    env:
      IS_DATAFLOW_END: true
      WRITE_LOG: true
```

## Important Rules
1. ONLY use agents from the "Available Agents" list above
2. Use the exact input/output parameter names shown for each agent
3. Start with terminal-input node for user input
4. Connect outputs to inputs: format is `node-id/output-name`
5. Last node MUST have `IS_DATAFLOW_END: true` and `WRITE_LOG: true` in env
6. Build command format: `pip install -e ../../agents/AGENT_NAME`
7. Path should match the agent name
8. Keep it simple (2-4 agents maximum)

Output ONLY the YAML content, no explanations or markdown:
"""

        yaml_str = client.generate(prompt)

        # Clean up markdown
        yaml_str = re.sub(r'```yaml\n', '', yaml_str)
        yaml_str = re.sub(r'```\n?', '', yaml_str)

        return yaml_str.strip()

    def _extract_agent_names(self, flow_yaml: str) -> List[str]:
        """Extract agent names from flow YAML"""
        import yaml

        try:
            flow_data = yaml.safe_load(flow_yaml)
            if 'nodes' in flow_data:
                return [
                    node.get('id', '')
                    for node in flow_data['nodes']
                    if node.get('id') != 'terminal-input' and node.get('id')
                ]
        except:
            pass

        return []



"""
Model Manager for PrimeSpeech TTS.
Handles downloading and managing GPT-SoVITS models from HuggingFace.
Based on VoiceDialogue's model management system.
"""

import os
import json
import shutil
from pathlib import Path
from typing import Dict, Optional, List
from concurrent.futures import ThreadPoolExecutor

try:
    from huggingface_hub import hf_hub_download, HfFileSystem
    HF_HUB_AVAILABLE = True
except ImportError:
    HF_HUB_AVAILABLE = False
    print("Warning: huggingface_hub not installed. Model downloading will be limited.")
    print("Install with: pip install huggingface-hub")


# Base pretrained model files that are shared across all voices
BASE_PRETRAINED_FILES = {
    'chinese-hubert-base/config.json': 'chinese-hubert-base/config.json',
    'chinese-hubert-base/preprocessor_config.json': 'chinese-hubert-base/preprocessor_config.json', 
    'chinese-hubert-base/pytorch_model.bin': 'chinese-hubert-base/pytorch_model.bin',
    'chinese-roberta-wwm-ext-large/config.json': 'chinese-roberta-wwm-ext-large/config.json',
    'chinese-roberta-wwm-ext-large/pytorch_model.bin': 'chinese-roberta-wwm-ext-large/pytorch_model.bin',
    'chinese-roberta-wwm-ext-large/tokenizer.json': 'chinese-roberta-wwm-ext-large/tokenizer.json',
}


class ModelManager:
    """Manages GPT-SoVITS model downloads and storage."""
    
    def __init__(self, models_dir: Path):
        """Initialize model manager.
        
        Args:
            models_dir: Directory to store models
        """
        self.models_dir = Path(models_dir)
        self.moyoyo_dir = self.models_dir / "moyoyo"  # MoYoYo models directory
        self.moyoyo_dir.mkdir(parents=True, exist_ok=True)
        
        # Model repository on HuggingFace
        self.repository = "MoYoYoTech/tone-models"
        
    def check_models_exist(self, voice_name: str, voice_config: Dict, verbose: bool = True) -> bool:
        """Check if model files exist locally.
        
        Args:
            voice_name: Name of the voice
            voice_config: Voice configuration dictionary
            verbose: Whether to print missing files
            
        Returns:
            True if all required model files exist
        """
        # Check voice-specific files
        required_files = [
            voice_config.get("gpt_weights", ""),
            voice_config.get("sovits_weights", ""),
            voice_config.get("reference_audio", "")
        ]
        
        for file_name in required_files:
            if file_name:
                file_path = self.moyoyo_dir / file_name
                if not file_path.exists():
                    if verbose:
                        print(f"Missing: {file_path}")
                    return False
        
        # Check base pretrained models
        for file_path in BASE_PRETRAINED_FILES.keys():
            full_path = self.moyoyo_dir / file_path
            if not full_path.exists():
                if verbose:
                    print(f"Missing base model: {full_path}")
                return False
                
        return True
    
    def download_file_from_huggingface(self, filename: str, force: bool = False) -> Path:
        """Download a single file from HuggingFace.
        
        Args:
            filename: File path in the repository
            force: Force re-download even if exists
            
        Returns:
            Path to downloaded file
        """
        output_path = self.moyoyo_dir / filename
        
        # Check if file exists and skip if not forcing
        if output_path.exists() and not force:
            return output_path
        
        # Create parent directory
        output_path.parent.mkdir(parents=True, exist_ok=True)
        
        if HF_HUB_AVAILABLE:
            try:
                print(f"Downloading: {filename}")
                hf_hub_download(
                    repo_id=self.repository,
                    filename=filename,
                    local_dir=str(self.moyoyo_dir),
                    cache_dir=str(self.moyoyo_dir / ".cache")
                )
                print(f"✓ Downloaded: {filename}")
            except Exception as e:
                print(f"Failed to download {filename}: {e}")
                # Don't raise - allow fallback to placeholder
        else:
            print(f"Cannot download {filename}: huggingface_hub not installed")
            
        return output_path
    
    def get_voice_model_paths(self, voice_name: str, voice_config: Dict) -> Dict[str, Path]:
        """Get or download model files for a voice.
        
        Args:
            voice_name: Name of the voice
            voice_config: Voice configuration dictionary
            
        Returns:
            Dictionary of model file paths
        """
        # Download models if needed
        if not self.check_models_exist(voice_name, voice_config):
            self.download_voice_models(voice_name, voice_config)
        
        # Return paths
        return {
            "gpt_model": self.moyoyo_dir / voice_config.get("gpt_weights", "GPT_weights/placeholder.ckpt"),
            "sovits_model": self.moyoyo_dir / voice_config.get("sovits_weights", "SoVITS_weights/placeholder.pth"),
            "reference_audio": self.moyoyo_dir / voice_config.get("reference_audio", "ref_audios/placeholder.wav"),
            "chinese-hubert-base": self.moyoyo_dir / "chinese-hubert-base",
            "chinese-roberta-wwm-ext-large": self.moyoyo_dir / "chinese-roberta-wwm-ext-large",
        }
    
    def download_voice_models(self, voice_name: str, voice_config: Dict):
        """Download models for a specific voice.
        
        Args:
            voice_name: Name of the voice
            voice_config: Voice configuration dictionary
        """
        print(f"Downloading models for {voice_name}...")
        
        if not HF_HUB_AVAILABLE:
            print("Warning: Creating placeholder model files (huggingface_hub not available)")
            self._create_placeholder_models(voice_name, voice_config)
            return
        
        # Files to download
        files_to_download = []
        
        # Add voice-specific files
        for key in ["gpt_weights", "sovits_weights", "reference_audio"]:
            if key in voice_config:
                files_to_download.append(voice_config[key])
        
        # Add base pretrained models
        files_to_download.extend(BASE_PRETRAINED_FILES.keys())
        
        # Try to download files (don't fail if some are missing)
        downloaded_count = 0
        for filename in files_to_download:
            if filename:
                try:
                    self.download_file_from_huggingface(filename)
                    downloaded_count += 1
                except Exception as e:
                    print(f"Skipping {filename}: {e}")
        
        # If no files were downloaded, create placeholders
        if downloaded_count == 0:
            print("No files downloaded, creating placeholders...")
            self._create_placeholder_models(voice_name, voice_config)
        else:
            print(f"✓ Downloaded {downloaded_count} files for {voice_name}")
    
    def _create_placeholder_models(self, voice_name: str, voice_config: Dict):
        """Create placeholder model files for testing.
        
        Args:
            voice_name: Name of the voice
            voice_config: Voice configuration dictionary
        """
        # Create placeholder voice files
        placeholders = [
            (voice_config.get("gpt_weights", "GPT_weights/placeholder.ckpt"), b"GPT_PLACEHOLDER"),
            (voice_config.get("sovits_weights", "SoVITS_weights/placeholder.pth"), b"SOVITS_PLACEHOLDER"),
            (voice_config.get("reference_audio", "ref_audios/placeholder.wav"), b"AUDIO_PLACEHOLDER"),
        ]
        
        for file_name, content in placeholders:
            if file_name:
                file_path = self.moyoyo_dir / file_name
                file_path.parent.mkdir(parents=True, exist_ok=True)
                if not file_path.exists():
                    file_path.write_bytes(content)
        
        # Create placeholder base models
        for file_path in BASE_PRETRAINED_FILES.keys():
            full_path = self.moyoyo_dir / file_path
            full_path.parent.mkdir(parents=True, exist_ok=True)
            if not full_path.exists():
                full_path.write_bytes(b"BASE_MODEL_PLACEHOLDER")
        
        print(f"Created placeholder models for {voice_name}")
    
    def get_model_size(self, voice_name: str, voice_config: Dict) -> float:
        """Get the total size of downloaded model files for a voice.
        
        Args:
            voice_name: Name of the voice
            voice_config: Voice configuration dictionary
            
        Returns:
            Total size in MB
        """
        total_size = 0
        
        # Check voice-specific files
        files_to_check = [
            self.moyoyo_dir / voice_config.get("gpt_weights", ""),
            self.moyoyo_dir / voice_config.get("sovits_weights", ""),
            self.moyoyo_dir / voice_config.get("reference_audio", "")
        ]
        
        for file_path in files_to_check:
            if file_path and file_path.exists():
                total_size += file_path.stat().st_size
        
        # Convert to MB
        return total_size / (1024 * 1024)
    
    def list_available_voices(self) -> Dict[str, Dict]:
        """List all downloaded voice models.
        
        Returns:
            Dictionary mapping voice names to their metadata
        """
        voices = {}
        
        # Import VOICE_CONFIGS to check against
        from .config import VOICE_CONFIGS
        
        # Check which voices have all required files downloaded
        for voice_name, voice_config in VOICE_CONFIGS.items():
            if self.check_models_exist(voice_name, voice_config, verbose=False):
                voices[voice_name] = {
                    "repository": voice_config.get("repository", self.repository),
                    "language": voice_config.get("text_lang", "unknown"),
                    "size_mb": self.get_model_size(voice_name, voice_config)
                }
        
        return voices
    
    def check_file_exists_on_huggingface(self, filename: str) -> bool:
        """Check if a file exists on HuggingFace.
        
        Args:
            filename: File path to check
            
        Returns:
            True if file exists on HuggingFace
        """
        if not HF_HUB_AVAILABLE:
            return False
            
        try:
            fs = HfFileSystem()
            remote_files = fs.ls(f'{self.repository}/{filename}')
            return len(remote_files) > 0
        except:
            return False
    
    def get_model_info(self) -> Dict:
        """Get information about available models.
        
        Returns:
            Dictionary with model information
        """
        info = {
            "repository": self.repository,
            "models_dir": str(self.moyoyo_dir),
            "downloaded_voices": self.list_available_voices(),
            "huggingface_available": HF_HUB_AVAILABLE,
        }
        
        # Check disk usage
        total_size = 0
        if self.moyoyo_dir.exists():
            for file_path in self.moyoyo_dir.rglob("*"):
                if file_path.is_file():
                    total_size += file_path.stat().st_size
        
        info["total_size_mb"] = total_size / (1024 * 1024)
        
        return info


import os
import sys
from PyQt5.QtCore import QEvent
from PyQt5.QtWidgets import QApplication, QMainWindow, QLabel, QLineEdit, QPushButton, QTextEdit
from PyQt5.QtWidgets import QGridLayout, QVBoxLayout, QWidget, QFileDialog, QStatusBar, QComboBox
import soundfile as sf

from tools.i18n.i18n import I18nAuto
i18n = I18nAuto()

from inference_webui import gpt_path, sovits_path, change_gpt_weights, change_sovits_weights, get_tts_wav


class GPTSoVITSGUI(QMainWindow):
    GPT_Path = gpt_path
    SoVITS_Path = sovits_path

    def __init__(self):
        super().__init__()

        self.setWindowTitle('GPT-SoVITS GUI')
        self.setGeometry(800, 450, 950, 850)

        self.setStyleSheet("""
            QWidget {
                background-color: #a3d3b1; 
            }

            QTabWidget::pane {
                background-color: #a3d3b1;  
            }

            QTabWidget::tab-bar {
                alignment: left;
            }

            QTabBar::tab {
                background: #8da4bf; 
                color: #ffffff;  
                padding: 8px;
            }

            QTabBar::tab:selected {
                background: #2a3f54; 
            }

            QLabel {
                color: #000000;  
            }

            QPushButton {
                background-color: #4CAF50; 
                color: white;  
                padding: 8px;
                border: 1px solid #4CAF50;
                border-radius: 4px;
            }

            QPushButton:hover {
                background-color: #45a049;  
                border: 1px solid #45a049;
                box-shadow: 2px 2px 2px rgba(0, 0, 0, 0.1);
            }
        """)    

        license_text = (
        "本软件以MIT协议开源, 作者不对软件具备任何控制力, 使用软件者、传播软件导出的声音者自负全责. "
        "如不认可该条款, 则不能使用或引用软件包内任何代码和文件. 详见根目录LICENSE.")
        license_label = QLabel(license_text)
        license_label.setWordWrap(True)

        self.GPT_model_label = QLabel("选择GPT模型:")
        self.GPT_model_input = QLineEdit()
        self.GPT_model_input.setPlaceholderText("拖拽或选择文件")
        self.GPT_model_input.setText(self.GPT_Path)
        self.GPT_model_input.setReadOnly(True)
        self.GPT_model_button = QPushButton("选择GPT模型文件")
        self.GPT_model_button.clicked.connect(self.select_GPT_model)

        self.SoVITS_model_label = QLabel("选择SoVITS模型:")
        self.SoVITS_model_input = QLineEdit()
        self.SoVITS_model_input.setPlaceholderText("拖拽或选择文件")
        self.SoVITS_model_input.setText(self.SoVITS_Path)
        self.SoVITS_model_input.setReadOnly(True)
        self.SoVITS_model_button = QPushButton("选择SoVITS模型文件")
        self.SoVITS_model_button.clicked.connect(self.select_SoVITS_model)

        self.ref_audio_label = QLabel("上传参考音频:")
        self.ref_audio_input = QLineEdit()
        self.ref_audio_input.setPlaceholderText("拖拽或选择文件")
        self.ref_audio_input.setReadOnly(True)
        self.ref_audio_button = QPushButton("选择音频文件")
        self.ref_audio_button.clicked.connect(self.select_ref_audio)

        self.ref_text_label = QLabel("参考音频文本:")
        self.ref_text_input = QLineEdit()
        self.ref_text_input.setPlaceholderText("直接输入文字或上传文本")
        self.ref_text_button = QPushButton("上传文本")
        self.ref_text_button.clicked.connect(self.upload_ref_text)

        self.ref_language_label = QLabel("参考音频语言:")
        self.ref_language_combobox = QComboBox()
        self.ref_language_combobox.addItems(["中文", "英文", "日文", "中英混合", "日英混合", "多语种混合"])
        self.ref_language_combobox.setCurrentText("多语种混合")

        self.target_text_label = QLabel("合成目标文本:")
        self.target_text_input = QLineEdit()
        self.target_text_input.setPlaceholderText("直接输入文字或上传文本")
        self.target_text_button = QPushButton("上传文本")
        self.target_text_button.clicked.connect(self.upload_target_text)

        self.target_language_label = QLabel("合成音频语言:")
        self.target_language_combobox = QComboBox()
        self.target_language_combobox.addItems(["中文", "英文", "日文", "中英混合", "日英混合", "多语种混合"])
        self.target_language_combobox.setCurrentText("多语种混合")

        self.output_label = QLabel("输出音频路径:")
        self.output_input = QLineEdit()
        self.output_input.setPlaceholderText("拖拽或选择文件")
        self.output_input.setReadOnly(True)
        self.output_button = QPushButton("选择文件夹")
        self.output_button.clicked.connect(self.select_output_path)

        self.output_text = QTextEdit()
        self.output_text.setReadOnly(True)

        self.add_drag_drop_events([
            self.GPT_model_input,
            self.SoVITS_model_input,
            self.ref_audio_input,
            self.ref_text_input,
            self.target_text_input,
            self.output_input,
        ])

        self.synthesize_button = QPushButton("合成")
        self.synthesize_button.clicked.connect(self.synthesize)

        self.clear_output_button = QPushButton("清空输出")
        self.clear_output_button.clicked.connect(self.clear_output)

        self.status_bar = QStatusBar()

        main_layout = QVBoxLayout()

        input_layout = QGridLayout(self)
        input_layout.setSpacing(10)

        input_layout.addWidget(license_label, 0, 0, 1, 3)

        input_layout.addWidget(self.GPT_model_label, 1, 0)
        input_layout.addWidget(self.GPT_model_input, 2, 0, 1, 2)
        input_layout.addWidget(self.GPT_model_button, 2, 2)

        input_layout.addWidget(self.SoVITS_model_label, 3, 0)
        input_layout.addWidget(self.SoVITS_model_input, 4, 0, 1, 2)
        input_layout.addWidget(self.SoVITS_model_button, 4, 2)

        input_layout.addWidget(self.ref_audio_label, 5, 0)
        input_layout.addWidget(self.ref_audio_input, 6, 0, 1, 2)
        input_layout.addWidget(self.ref_audio_button, 6, 2)

        input_layout.addWidget(self.ref_language_label, 7, 0)
        input_layout.addWidget(self.ref_language_combobox, 8, 0, 1, 1)
        input_layout.addWidget(self.ref_text_label, 9, 0)
        input_layout.addWidget(self.ref_text_input, 10, 0, 1, 2)
        input_layout.addWidget(self.ref_text_button, 10, 2)

        input_layout.addWidget(self.target_language_label, 11, 0)
        input_layout.addWidget(self.target_language_combobox, 12, 0, 1, 1)
        input_layout.addWidget(self.target_text_label, 13, 0)
        input_layout.addWidget(self.target_text_input, 14, 0, 1, 2)
        input_layout.addWidget(self.target_text_button, 14, 2)

        input_layout.addWidget(self.output_label, 15, 0)
        input_layout.addWidget(self.output_input, 16, 0, 1, 2)
        input_layout.addWidget(self.output_button, 16, 2)

        main_layout.addLayout(input_layout)

        output_layout = QVBoxLayout()
        output_layout.addWidget(self.output_text)
        main_layout.addLayout(output_layout)

        main_layout.addWidget(self.synthesize_button)

        main_layout.addWidget(self.clear_output_button)

        main_layout.addWidget(self.status_bar)

        self.central_widget = QWidget()
        self.central_widget.setLayout(main_layout)
        self.setCentralWidget(self.central_widget)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.acceptProposedAction()

    def dropEvent(self, event):
        if event.mimeData().hasUrls():
            file_paths = [url.toLocalFile() for url in event.mimeData().urls()]
            if len(file_paths) == 1:
                self.update_ref_audio(file_paths[0])
            else:
                self.update_ref_audio(", ".join(file_paths))

    def add_drag_drop_events(self, widgets):
        for widget in widgets:
            widget.setAcceptDrops(True)
            widget.installEventFilter(self)

    def eventFilter(self, obj, event):
        if event.type() in (QEvent.DragEnter, QEvent.Drop):
            mime_data = event.mimeData()
            if mime_data.hasUrls():
                event.acceptProposedAction()

        return super().eventFilter(obj, event)

    def select_GPT_model(self):
        file_path, _ = QFileDialog.getOpenFileName(self, "选择GPT模型文件", "", "GPT Files (*.ckpt)")
        if file_path:
            self.GPT_model_input.setText(file_path)

    def select_SoVITS_model(self):
        file_path, _ = QFileDialog.getOpenFileName(self, "选择SoVITS模型文件", "", "SoVITS Files (*.pth)")
        if file_path:
            self.SoVITS_model_input.setText(file_path)

    def select_ref_audio(self):
        file_path, _ = QFileDialog.getOpenFileName(self, "选择参考音频文件", "", "Audio Files (*.wav *.mp3)")
        if file_path:
            self.update_ref_audio(file_path)

    def upload_ref_text(self):
        file_path, _ = QFileDialog.getOpenFileName(self, "选择文本文件", "", "Text Files (*.txt)")
        if file_path:
            with open(file_path, 'r', encoding='utf-8') as file:
                content = file.read()
                self.ref_text_input.setText(content)

    def upload_target_text(self):
        file_path, _ = QFileDialog.getOpenFileName(self, "选择文本文件", "", "Text Files (*.txt)")
        if file_path:
            with open(file_path, 'r', encoding='utf-8') as file:
                content = file.read()
                self.target_text_input.setText(content)

    def select_output_path(self):
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        options |= QFileDialog.ShowDirsOnly

        folder_dialog = QFileDialog()
        folder_dialog.setOptions(options)
        folder_dialog.setFileMode(QFileDialog.Directory)

        if folder_dialog.exec_():
            folder_path = folder_dialog.selectedFiles()[0]
            self.output_input.setText(folder_path)

    def update_ref_audio(self, file_path):
        self.ref_audio_input.setText(file_path)

    def clear_output(self):
        self.output_text.clear()

    def synthesize(self):
        GPT_model_path = self.GPT_model_input.text()
        SoVITS_model_path = self.SoVITS_model_input.text()
        ref_audio_path = self.ref_audio_input.text()
        language_combobox = self.ref_language_combobox.currentText()
        language_combobox = i18n(language_combobox)
        ref_text = self.ref_text_input.text()
        target_language_combobox = self.target_language_combobox.currentText()
        target_language_combobox = i18n(target_language_combobox)
        target_text = self.target_text_input.text()
        output_path = self.output_input.text()

        if GPT_model_path != self.GPT_Path:
            change_gpt_weights(gpt_path=GPT_model_path)
            self.GPT_Path = GPT_model_path
        if SoVITS_model_path != self.SoVITS_Path:
            change_sovits_weights(sovits_path=SoVITS_model_path)
            self.SoVITS_Path = SoVITS_model_path

        synthesis_result = get_tts_wav(ref_wav_path=ref_audio_path, 
                                       prompt_text=ref_text, 
                                       prompt_language=language_combobox, 
                                       text=target_text, 
                                       text_language=target_language_combobox)

        result_list = list(synthesis_result)

        if result_list:
            last_sampling_rate, last_audio_data = result_list[-1]
            output_wav_path = os.path.join(output_path, "output.wav") 
            sf.write(output_wav_path, last_audio_data, last_sampling_rate)

            result = "Audio saved to " + output_wav_path

        self.status_bar.showMessage("合成完成！输出路径：" + output_wav_path, 5000)
        self.output_text.append("处理结果：\n" + result)


if __name__ == '__main__':
    app = QApplication(sys.argv)
    mainWin = GPTSoVITSGUI()
    mainWin.show()
    sys.exit(app.exec_())


import copy
import json
import os
from functools import wraps
import traceback
from os import mkdir
import pyarrow as pa
from attrs import define, field
from typing import Any, Dict, Union
import logging

from mcp.server import FastMCP

from mofa.utils.files.read import read_yaml
import yaml
from dora import Node
from dotenv import load_dotenv
from pathlib import Path
from mofa.utils.files.write import ensure_directory_exists
from logging.handlers import RotatingFileHandler

def create_agent_output(agent_name:str, agent_result:Union[str,dict,list], dataflow_status:bool=os.getenv(
                                                                                    "IS_DATAFLOW_END",                                                              True)):
    if isinstance(agent_result, dict) or isinstance(agent_result, list):
        agent_result = json.dumps(agent_result, ensure_ascii=False)
    return json.dumps({'step_name':agent_name, 'node_results':agent_result, 'dataflow_status':dataflow_status}, ensure_ascii=False)
def load_node_result(node_data):
    return json.loads(node_data).get('node_results')
@define
class MofaLogger:
    agent_name: str
    log_dir: str = field(default='logs')
    log_file: str = field(default='agent.log')
    max_log_size: int = field(default=10*1024*1024)  # 默认10MB
    backup_count: int = field(default=5)
    logger: logging.Logger = field(init=False, default=None)

    def __attrs_post_init__(self):
        """
        初始化 NodeLogger 实例。
        """
        self.logger = logging.getLogger(self.agent_name)
        self._setup_logger()

        if os.getenv('LOG_FILE', None) is not None:
            self.log_file = os.getenv('LOG_FILE')
            ensure_directory_exists(file_path=self.log_file)

    def _setup_logger(self):
        """
        设置日志记录器，包括日志目录的创建和日志处理器的配置。
        """
        # 创建日志目录（如果不存在）
        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir, exist_ok=True)

        # 创建日志文件的完整路径
        log_path = os.path.join(self.log_dir, self.log_file)

        # 创建一个 RotatingFileHandler，用于日志轮转
        handler = RotatingFileHandler(
            log_path,
            maxBytes=self.max_log_size,
            backupCount=self.backup_count,
            encoding='utf-8'
        )
        handler.setLevel(logging.DEBUG)

        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)

        # 将处理器添加到记录器
        self.logger.addHandler(handler)
        self.logger.setLevel(logging.DEBUG)

    def log(self,message:str, level:str='INFO' ):
        """
        记录日志消息。

        :param level: 日志级别，例如 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'
        :param message: 要记录的日志消息
        """
        log_method = getattr(self.logger, level.lower(), None)
        if log_method:
            log_method(message)
        else:
            self.logger.error(f"Invalid log level: {level}. Message: {message}")


@define
class MofaAgent:
    agent_name:str = field(default='mofa-agent')
    node:Any = field(default=None)
    agent_inputs:dict = field(factory=dict)
    event:Any = field(factory=dict)
    event_time_out:int = field(default=20)
    is_write_log:bool = field(default=False)
    log_file:str=field(default='agent.log')
    agent_log:MofaLogger = field(init=False)
    mcp :Any = field(default=None)
    def __attrs_post_init__(self):
        self.node = Node(self.agent_name)
        env_files = ['.env.secret', '.env']
        for env_file in env_files:
            if os.path.exists(env_file):
                load_dotenv(dotenv_path=env_file)
        log_params = ['IS_WRITE_LOG', 'WRITE_LOG']
        for log_status in log_params:
            if os.getenv(log_status, None) is not None:
                self.is_write_log = os.getenv(log_status)
        self.agent_log = MofaLogger(agent_name=self.agent_name, log_file=self.log_file)

    def __init_mcp(self):
        if os.getenv('MCP', None) is not None and self.mcp is None:
            self.mcp = FastMCP(self.agent_name)

    def register_mcp_tool(self, func):
        """动态注册 MCP 工具"""
        tool_name = func.__name__
        self.__init_mcp()
        decorated_func = self.mcp.tool()(func)
        # setattr(self, tool_name, decorated_func)
        print(f"工具 '{tool_name}' 注册成功。")

    def _parse_event_value(self, event):
        try:
            return load_node_result(event["value"][0].as_py())
        except Exception:
            return event["value"][0].as_py()

    def _receive_event_input(self, event, parameter_names:Union[str,dict]):
        if event["type"] == "INPUT":
            if isinstance(parameter_names,str):

                if event['id'] == parameter_names :
                    input_data = self._parse_event_value(event=event)
                    return input_data
            elif isinstance(parameter_names,dict):
                data = copy.deepcopy(parameter_names)
                if event['id'] in list(data.keys()) :
                    data[event['id']] = self._parse_event_value(event=event)
                    return data
                else:
                    return data
    def receive_parameter(self,parameter_name:str):
        for event in self.node:
            input_data = self._receive_event_input(event=event, parameter_names=parameter_name)
            if input_data is not None:
                self.event = event
                self.write_log(message=json.dumps(f"{self.agent_name}  receive  data : {input_data}  "))
                return input_data
            else:
                continue
            # self.node.next(self.event_time_out)

    def receive_parameters(self,parameter_names:list)->dict:
        parameter_data = {}
        if len(parameter_names) > 0:
            parameter_data = {key: None for key in parameter_names}
        for event in self.node:

            parameter_data = self._receive_event_input(event=event,parameter_names=parameter_data)
            self.event = event
            # self.node.next(self.event_time_out)
            is_parameter_data_status = all(value is not None for value in parameter_data.values())
            if is_parameter_data_status :
                break
        self.write_log(message=json.dumps(f"{self.agent_name}  receive parameters data : {parameter_data}  "))
        return parameter_data
            
    def send_output(self, agent_output_name: str, agent_result: Any, is_end_status=os.getenv('IS_DATAFLOW_END', True)):
        if is_end_status == 'true' or is_end_status == 'True':
            is_end_status = True
        self.node.send_output(
            agent_output_name,
            pa.array([create_agent_output(
                agent_name=agent_output_name,
                agent_result=agent_result,
                dataflow_status=is_end_status
            )]),
            self.event['metadata']
        )
        if agent_result == "None" or agent_result == " " or agent_result == "" or agent_result is None or agent_result == [] or agent_result == '[]':
            return
        self.write_log(message=json.dumps(f"{agent_output_name}  output data : {agent_result}  type : {type(agent_result)}" ))

    def write_log(self, message:str, level:str='INFO'):
        if self.is_write_log:
            if message == "None" or message == " " or message == "" or message is None or message == [] or message == '[]':
                return
            else:
                self.agent_log.log(message=message, level=level)
    def run_mcp(self,mcp_transport:str='sse'):
        if self.mcp is not None:
            print('mcp server 运行成功')
            self.mcp.run(transport=mcp_transport)


@define
class BaseMofaAgent:
    """
    A base class for MOFA agents, providing configuration management,
    resource initialization, and a context manager interface.
    """

    config_path: Union[str, None] = field(default=None)
    config: Dict[str, Any] = field(init=False,factory=dict)
    llm_client: Any = field(init=False, default=None)
    llm_config_path: Union[str, None] = field(default='.env.secret')
    
    def __attrs_post_init__(self) -> None:
        """
        Automatically called by attrs after the object is instantiated.
        Loads the configuration, initializes the client, 
        and sets up additional resources.
        """
        if self.config_path is None:
            self.config_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), ) + '/configs/agent.yml'
        
    def init_llm_config(self,llm_config_path:str=None):
        if llm_config_path is None :
            llm_config_path = self.llm_config_path
        if Path(llm_config_path).exists():
            load_dotenv(llm_config_path)

    def load_config(self, config_path: str=None) -> Dict[str, Any]:
        """
        Load configuration from a YAML file.
        """
        if config_path is None:
            config_path = self.config_path
        return read_yaml(file_path = config_path)

    def send_output(self, node:Node,event,output_name: str, output_data: Any,output_step_name:str=None,is_end_dataflow:bool=True):
        if output_step_name is None:
            output_step_name = output_name
        node.send_output(output_name, pa.array([create_agent_output(agent_name=output_step_name, agent_result=output_data, dataflow_status=os.getenv('IS_DATAFLOW_END', is_end_dataflow))]), event['metadata'])
    
    def run(self,task:str=None,*args,**kwargs):
        pass
    
    def parse_agent_parameters(self,agent_input:Any):
        try:
            agent_input = load_node_result(agent_input)
        except:
            agent_input = agent_input
        return agent_input
    
    def create_llm_client(self,config:dict=None,*args,**kwargs):
        if config is None:
            config = self.config
        pass


def run_agent(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        while True:
            try:
                func(*args, **kwargs)
            except Exception as e:
                print(f"Error occurred: {e}")
                traceback.print_exc()
    return wrapper



import click
import ast
from typing import List, Dict

# def parse_value(value_str: str):
#     """解析值为Python原生类型（支持字符串、列表、字典等）"""
#     try:
#         # 尝试解析为Python字面量（处理列表、字典等）
#         return ast.literal_eval(value_str)
#     except (SyntaxError, ValueError):
#         # 解析失败则作为原始字符串返回
#         return value_str

def collect_interactive_input(unit_test=True) -> List[Dict]:
    """support collect test cases interactively or user input mode."""
    if unit_test:
        test_cases = []
        click.echo("===== 交互式测试用例输入 =====")
        
        case_index = 1  # 用例序号，用于默认名称
        while True:
            case = {}
            
            # 1. 用例名称（默认自动生成，可自定义）
            default_name = f"test_case_{case_index}"
            case_name = click.prompt(
                f"请输入用例名称（默认: {default_name}）",
                type=str,
                default=default_name,
                show_default=False
            )
            case["name"] = case_name
            
            # 2. 输入参数（input）：支持key:value或key:[...]格式
            click.echo("\n请输入input参数")
            click.echo("  普通值如： hello")
            click.echo("  列表如： [\"a\", \"b\", \"c\"]")
            input_str = click.prompt("input参数")
            input_dict = {}
            
            item = input_str.strip()
            if not item:
                continue
            case["input"] = item
            
            # 3. 预期输出（expected_output）：同上支持多种类型
            click.echo("\n请输入预期输出")
            click.echo("  普通值如：hello ")
            click.echo("  列表如： [\"a\", \"b\", \"c\"]")
            expected_output_str = click.prompt("expected_output参数")
            
            item = expected_output_str.strip()
            if not item:
                continue
            case["expected_output"] = item
            
            test_cases.append(case)
            case_index += 1
            
            # Whether to continue adding more test cases
            if not click.confirm("\n是否继续添加下一个测试用例？", default=False):
                break
        
        click.echo(f"\n已收集 {len(test_cases)} 个测试用例，开始执行...")
        return test_cases
    else:
        # In non-unit-test mode, just collect a single input from user
        test_cases = []
        click.echo("===== 运行节点，输入参数 =====")
        input_str = click.prompt("请输入input参数")

        item = input_str.strip()
        if not item:
            raise click.BadParameter("输入不能为空")

        case = {
            "name": "user_input",
            "input": item
        }
        return [case]



"""
Simplified Dora PrimeSpeech Node - Main entry point
High-quality text-to-speech using GPT-SoVITS technology.
"""

import time
import os
import sys
import traceback
import json
import numpy as np
import pyarrow as pa
try:
    from mofa.agent_build.base.base_agent import MofaAgent, run_agent
    MOFA_AVAILABLE = True
except ImportError:
    from dora import Node
    MOFA_AVAILABLE = False
from pathlib import Path
from typing import Optional

from .config import PrimeSpeechConfig, VOICE_CONFIGS
from .model_manager import ModelManager
from .moyoyo_tts_wrapper_streaming_fix import StreamingMoYoYoTTSWrapper as MoYoYoTTSWrapper, MOYOYO_AVAILABLE


def send_log(node, level, message, config_level="INFO"):
    """Send log message through log output channel."""
    LOG_LEVELS = {
        "DEBUG": 10,
        "INFO": 20,
        "WARNING": 30,
        "ERROR": 40
    }
    
    if LOG_LEVELS.get(level, 0) < LOG_LEVELS.get(config_level, 20):
        return
    
    formatted_message = f"[{level}] {message}"
    # Also print to console so errors show in docker logs
    try:
        print(formatted_message, file=sys.stderr if level in {"ERROR", "WARNING"} else sys.stdout, flush=True)
    except Exception:
        pass
    log_data = {
        "node": "primespeech",
        "level": level,
        "message": formatted_message,
        "timestamp": time.time()
    }
    node.send_output("log", pa.array([json.dumps(log_data)]))


def validate_language_config(lang_code, param_name, node, log_level):
    """Validate language configuration and provide helpful error messages"""
    # Valid language codes for MoYoYo TTS v2
    VALID_LANGUAGES = ["auto", "auto_yue", "en", "zh", "ja", "yue", "ko",
                      "all_zh", "all_ja", "all_yue", "all_ko"]

    if lang_code in VALID_LANGUAGES:
        return lang_code

    # Invalid language code - show error prominently
    error_header = "=" * 70
    print(f"\n{error_header}", flush=True)
    print(f"❌ PRIMESPEECH CONFIGURATION ERROR", flush=True)
    print(f"{error_header}", flush=True)

    main_error = f"INVALID {param_name}: '{lang_code}' is NOT a valid language!"
    print(f"{main_error}", flush=True)
    send_log(node, "ERROR", main_error, log_level)

    # Check for common mistakes and suggest corrections
    if lang_code.lower() == "cn":
        hint = "Did you mean 'zh' for Chinese? Use 'zh' not 'cn'!"
        print(f"💡 HINT: {hint}", flush=True)
        send_log(node, "ERROR", hint, log_level)
    elif lang_code.lower() == "chinese":
        hint = "Use 'zh' for Chinese, not 'chinese'!"
        print(f"💡 HINT: {hint}", flush=True)
        send_log(node, "ERROR", hint, log_level)
    elif lang_code.lower() == "english":
        hint = "Use 'en' for English, not 'english'!"
        print(f"💡 HINT: {hint}", flush=True)
        send_log(node, "ERROR", hint, log_level)

    valid_msg = f"Valid languages: {', '.join(VALID_LANGUAGES)}"
    print(f"✅ {valid_msg}", flush=True)
    send_log(node, "ERROR", valid_msg, log_level)

    print(f"⚠️  TTS will FAIL until you fix {param_name} in your configuration!", flush=True)
    print(f"{error_header}\n", flush=True)
    send_log(node, "ERROR", f"TTS will fail until you fix {param_name}!", log_level)

    # Return the invalid code as-is (will cause TTS to fail with clear error)
    return lang_code


def _validate_models_path(logger, models_env_var="PRIMESPEECH_MODEL_DIR") -> Optional[Path]:
    """Validate that required model directory exists and contains MoYoYo subdir.
    Returns the resolved path if valid, else None.
    """
    raw = os.environ.get(models_env_var)
    if not raw:
        logger("ERROR", f"Missing {models_env_var} environment variable; TTS cannot load models")
        return None
    # Expand env vars (e.g., $HOME) and user (~)
    base = Path(os.path.expanduser(os.path.expandvars(raw)))
    if not base.exists():
        logger("ERROR", f"{models_env_var} points to non-existent path: {base}")
        return None
    moyoyo_dir = base / "moyoyo"
    if not moyoyo_dir.exists():
        logger("WARNING", f"Expected models under: {moyoyo_dir} (directory missing)")
    return base


def main():
    """Main entry point for PrimeSpeech node"""

    if MOFA_AVAILABLE:
        agent = MofaAgent(agent_name="primespeech", is_write_log=True)
        node = agent.node
    else:
        node = Node()
    config = PrimeSpeechConfig()

    # Get voice configuration
    voice_name = config.VOICE_NAME
    if voice_name not in VOICE_CONFIGS:
        send_log(node, "ERROR", f"Unknown voice: {voice_name}. Available: {list(VOICE_CONFIGS.keys())}", config.LOG_LEVEL)
        voice_name = "Doubao"

    voice_config = VOICE_CONFIGS[voice_name]

    # Override with environment variables if provided
    if config.PROMPT_TEXT:
        voice_config["prompt_text"] = config.PROMPT_TEXT

    # Validate and set text language
    send_log(node, "DEBUG", f"TEXT_LANG from env: '{config.TEXT_LANG}'", config.LOG_LEVEL)
    if config.TEXT_LANG:
        validated_text_lang = validate_language_config(
            config.TEXT_LANG, "TEXT_LANG", node, config.LOG_LEVEL)
        voice_config["text_lang"] = validated_text_lang
        send_log(node, "DEBUG", f"Validated TEXT_LANG: '{validated_text_lang}'", config.LOG_LEVEL)

    # Validate and set prompt language
    send_log(node, "DEBUG", f"PROMPT_LANG from env: '{config.PROMPT_LANG}'", config.LOG_LEVEL)
    if config.PROMPT_LANG:
        validated_prompt_lang = validate_language_config(
            config.PROMPT_LANG, "PROMPT_LANG", node, config.LOG_LEVEL)
        voice_config["prompt_lang"] = validated_prompt_lang
        send_log(node, "DEBUG", f"Validated PROMPT_LANG: '{validated_prompt_lang}'", config.LOG_LEVEL)
    
    # Add inference parameters
    effective_speed_factor = (
        config.SPEED_FACTOR
        if config.SPEED_FACTOR is not None
        else voice_config.get("speed_factor", 1.0)
    )

    if config.SPEED_FACTOR is not None:
        send_log(
            node,
            "INFO",
            f"Overriding speed_factor via env to {effective_speed_factor}",
            config.LOG_LEVEL,
        )

    voice_config.update({
        "top_k": config.TOP_K,
        "top_p": config.TOP_P,
        "temperature": config.TEMPERATURE,
        "speed_factor": effective_speed_factor,
        "batch_size": config.BATCH_SIZE,
        "seed": config.SEED,
        "text_split_method": config.TEXT_SPLIT_METHOD,
        "split_bucket": config.SPLIT_BUCKET,
        "return_fragment": config.RETURN_FRAGMENT,
        "fragment_interval": config.FRAGMENT_INTERVAL,
        "use_gpu": config.USE_GPU,
        "device": config.DEVICE,
        "sample_rate": config.SAMPLE_RATE,
    })
    
    # Initialize model manager
    model_manager = ModelManager(config.get_models_dir())
    
    send_log(node, "INFO", "PrimeSpeech Node initialized", config.LOG_LEVEL)
    
    if MOYOYO_AVAILABLE:
        send_log(node, "INFO", "✓ MoYoYo TTS engine available", config.LOG_LEVEL)
    else:
        send_log(node, "WARNING", "⚠️  MoYoYo TTS not fully available", config.LOG_LEVEL)
    
    # Log the configuration being used
    send_log(node, "INFO", f"Voice: {voice_name}", config.LOG_LEVEL)
    send_log(node, "INFO", f"Text Language: {voice_config.get('text_lang', 'auto')} (configured: {config.TEXT_LANG})", config.LOG_LEVEL)
    send_log(node, "INFO", f"Prompt Language: {voice_config.get('prompt_lang', 'auto')} (configured: {config.PROMPT_LANG})", config.LOG_LEVEL)

    # Print to stdout for immediate visibility
    speed_factor_value = voice_config.get('speed_factor')
    print(f"[PRIMESPEECH SPEED_FACTOR] Voice: {voice_name}, Speed: {speed_factor_value}, Env Override: {config.SPEED_FACTOR_OVERRIDE}", flush=True)

    send_log(node, "INFO", f"Speed Factor: {speed_factor_value} (env override: {config.SPEED_FACTOR_OVERRIDE is not None})", config.LOG_LEVEL)
    send_log(node, "INFO", f"Device: {config.DEVICE}", config.LOG_LEVEL)

    # Validate the final configuration
    final_text_lang = voice_config.get('text_lang', 'auto')
    final_prompt_lang = voice_config.get('prompt_lang', 'auto')

    VALID_LANGUAGES = ["auto", "auto_yue", "en", "zh", "ja", "yue", "ko",
                      "all_zh", "all_ja", "all_yue", "all_ko"]

    if final_text_lang not in VALID_LANGUAGES:
        send_log(node, "ERROR",
                f"CRITICAL: text_lang '{final_text_lang}' is not valid! "
                f"This will cause TTS to fail. Please fix your configuration.",
                config.LOG_LEVEL)

    if final_prompt_lang not in VALID_LANGUAGES:
        send_log(node, "ERROR",
                f"CRITICAL: prompt_lang '{final_prompt_lang}' is not valid! "
                f"This will cause TTS to fail. Please fix your configuration.",
                config.LOG_LEVEL)
    
    # Initialize TTS engine
    tts_engine: Optional[MoYoYoTTSWrapper] = None
    model_loaded = False
    
    # Statistics
    total_syntheses = 0
    total_duration = 0
    
    for event in node:
        if event["type"] == "INPUT":
            input_id = event["id"]
            
            if input_id == "text":
                # Get text to synthesize
                text = event["value"][0].as_py()
                metadata = event.get("metadata", {})

                # DEBUG: Log what we received
                send_log(node, "DEBUG", f"RECEIVED text: '{text}' (len={len(text)}, repr={repr(text)}, type={type(text).__name__})", config.LOG_LEVEL)

                segment_index = metadata.get("segment_index", -1)

                # Skip if text is only punctuation or whitespace
                text_stripped = text.strip()
                if not text_stripped or all(c in '。！？.!?,，、；：""''（）【】《》\n\r\t ' for c in text_stripped):
                    send_log(node, "DEBUG", f"SKIPPED - text is only punctuation/whitespace: '{text}'", config.LOG_LEVEL)
                    # Send segment_complete without audio
                    node.send_output(
                        "segment_complete",
                        pa.array(["skipped"]),
                        metadata={}
                    )
                    continue

                send_log(node, "INFO", f"Processing segment {segment_index + 1} (len={len(text)})", config.LOG_LEVEL)
                
                # Load models if not loaded
                if not model_loaded:
                    send_log(node, "INFO", "Loading models for the first time...", config.LOG_LEVEL)
                    # Validate models directory early so failures are visible
                    _validate_models_path(lambda lvl, msg: send_log(node, lvl, msg, config.LOG_LEVEL))

                    try:
                        # Always use PRIMESPEECH_MODEL_DIR
                        send_log(node, "INFO", "Using PRIMESPEECH_MODEL_DIR for models...", config.LOG_LEVEL)
                        # Initialize TTS engine
                        # Convert voice name to lowercase and remove spaces for MoYoYo compatibility
                        moyoyo_voice = voice_name.lower().replace(" ", "")
                        device = "cuda" if config.USE_GPU and config.DEVICE.startswith("cuda") else "cpu"

                        enable_streaming = config.RETURN_FRAGMENT if hasattr(config, 'RETURN_FRAGMENT') else False

                        # Initialize TTS wrapper using PRIMESPEECH_MODEL_DIR
                        tts_engine = MoYoYoTTSWrapper(
                            voice=moyoyo_voice,
                            device=device,
                            enable_streaming=enable_streaming,
                            chunk_duration=0.3,
                            voice_config=voice_config,
                            logger_func=lambda level, msg: send_log(node, level, msg, config.LOG_LEVEL)
                        )

                        # Check if initialization succeeded
                        if tts_engine is None or not hasattr(tts_engine, 'tts') or tts_engine.tts is None:
                            send_log(node, "ERROR", "TTS engine initialization failed!", config.LOG_LEVEL)
                            send_log(node, "ERROR", "TTS wrapper exists but internal TTS is None", config.LOG_LEVEL)
                        else:
                            send_log(node, "INFO", "TTS engine initialized successfully", config.LOG_LEVEL)
                        model_loaded = True
                        send_log(node, "INFO", "TTS engine ready", config.LOG_LEVEL)
                    except Exception as init_err:
                        send_log(node, "ERROR", f"TTS init error: {init_err}", config.LOG_LEVEL)
                        send_log(node, "ERROR", f"Traceback: {traceback.format_exc()}", config.LOG_LEVEL)
                        # Mark as not loaded and send error completion without audio
                        model_loaded = False
                        node.send_output(
                            "segment_complete",
                            pa.array(["error"]),
                            metadata={
                                "session_id": session_id,
                                "request_id": request_id,
                                "segment_index": segment_index,
                                "error": str(init_err),
                                "error_stage": "init"
                            }
                        )
                        # Skip this event since we cannot synthesize
                        continue
                
                # Synthesize speech
                start_time = time.time()
                
                try:
                    # Check if TTS engine is available
                    if tts_engine is None:
                        send_log(node, "ERROR", "Cannot synthesize - TTS engine is None!", config.LOG_LEVEL)
                        raise RuntimeError("TTS engine not initialized")
                    
                    if hasattr(tts_engine, 'tts') and tts_engine.tts is None:
                        send_log(node, "ERROR", "Cannot synthesize - internal TTS is None!", config.LOG_LEVEL)
                        raise RuntimeError("Internal TTS engine not initialized")
                    
                    language = voice_config.get("text_lang", "zh")
                    speed = voice_config.get("speed_factor", 1.0)
                    
                    if hasattr(tts_engine, 'enable_streaming') and tts_engine.enable_streaming:
                        # Streaming synthesis
                        send_log(node, "INFO", "Using streaming synthesis...", config.LOG_LEVEL)
                        fragment_num = 0
                        total_audio_duration = 0
                        
                        for sample_rate, audio_fragment in tts_engine.synthesize_streaming(text, language=language, speed=speed):
                            fragment_num += 1
                            fragment_duration = len(audio_fragment) / sample_rate
                            total_audio_duration += fragment_duration

                            # Guard against empty fragments
                            if audio_fragment is None or len(audio_fragment) == 0:
                                send_log(node, "WARNING", f"Skipping empty audio fragment {fragment_num}", config.LOG_LEVEL)
                            else:
                                # Ensure type is float32 for consistency
                                if audio_fragment.dtype != np.float32:
                                    audio_fragment = audio_fragment.astype(np.float32)
                                node.send_output(
                                    "audio",
                                    pa.array([audio_fragment]),
                                    metadata={
                                        "segment_index": segment_index,
                                        "segments_remaining": metadata.get("segments_remaining", 0),
                                        "question_id": metadata.get("question_id", "default"),  # Pass through question_id
                                        "fragment_num": fragment_num,
                                        "sample_rate": sample_rate,
                                        "duration": fragment_duration,
                                        "is_streaming": True,
                                    }
                                )
                        
                        synthesis_time = time.time() - start_time
                        send_log(node, "INFO", f"Streamed {fragment_num} fragments, {total_audio_duration:.2f}s audio in {synthesis_time:.3f}s", config.LOG_LEVEL)
                        # If nothing was streamed, mark as error to avoid hanging clients
                        if fragment_num == 0:
                            raise RuntimeError("No audio fragments produced during streaming synthesis")
                        
                    else:
                        # Batch synthesis
                        sample_rate, audio_array = tts_engine.synthesize(text, language=language, speed=speed)
                        
                        synthesis_time = time.time() - start_time
                        audio_duration = len(audio_array) / sample_rate
                        if audio_array is None or len(audio_array) == 0:
                            raise RuntimeError("TTS returned empty audio array")
                        # Normalize dtype
                        if audio_array.dtype != np.float32:
                            audio_array = audio_array.astype(np.float32)
                        
                        total_syntheses += 1
                        total_duration += audio_duration
                        
                        send_log(node, "INFO", f"Synthesized: {audio_duration:.2f}s audio in {synthesis_time:.3f}s", config.LOG_LEVEL)
                        
                        # Send audio output with segment counting metadata
                        node.send_output(
                            "audio",
                            pa.array([audio_array]),
                            metadata={
                                "segment_index": segment_index,
                                "segments_remaining": metadata.get("segments_remaining", 0),
                                "question_id": metadata.get("question_id", "default"),  # Pass through question_id
                                "sample_rate": sample_rate,
                                "duration": audio_duration,
                                "is_streaming": False,
                            }
                        )
                    
                    # Send segment completion signal
                    node.send_output(
                        "segment_complete",
                        pa.array(["completed"]),
                        metadata={}
                    )
                    send_log(node, "INFO", f"Sent segment_complete for segment {segment_index + 1}", config.LOG_LEVEL)
                    
                except Exception as e:
                    error_details = traceback.format_exc()

                    # Check for specific language-related errors
                    if "assert text_lang" in str(e) or "assert prompt_lang" in str(e) or "AssertionError" in str(e.__class__.__name__):
                        send_log(node, "ERROR", "="*60, config.LOG_LEVEL)
                        send_log(node, "ERROR", "CRITICAL: Language configuration error detected!", config.LOG_LEVEL)
                        send_log(node, "ERROR", f"TEXT_LANG: '{language}' (from config: '{config.TEXT_LANG}')", config.LOG_LEVEL)
                        send_log(node, "ERROR", f"PROMPT_LANG: '{voice_config.get('prompt_lang', 'auto')}' (from config: '{config.PROMPT_LANG}')", config.LOG_LEVEL)
                        send_log(node, "ERROR", "Valid languages: auto, auto_yue, zh, en, ja, ko, yue, all_zh, all_ja, all_yue, all_ko", config.LOG_LEVEL)
                        send_log(node, "ERROR", "Common mistakes: 'cn' should be 'zh', 'chinese' should be 'zh'", config.LOG_LEVEL)
                        send_log(node, "ERROR", "Fix your configuration and restart!", config.LOG_LEVEL)
                        send_log(node, "ERROR", "="*60, config.LOG_LEVEL)

                    send_log(node, "ERROR", f"Synthesis error: {e}", config.LOG_LEVEL)
                    send_log(node, "ERROR", f"Traceback: {error_details}", config.LOG_LEVEL)
                    
                    # Do NOT send invalid audio on error; only notify completion with error
                    node.send_output(
                        "segment_complete",
                        pa.array(["error"]),
                        metadata={
                            "error": str(e),
                            "error_stage": "synthesis"
                        }
                    )
                    send_log(node, "ERROR", f"Sent error segment_complete for segment {segment_index + 1}", config.LOG_LEVEL)
            
            elif input_id == "control":
                # Handle control commands
                command = event["value"][0].as_py()
                
                if command == "reset":
                    send_log(node, "INFO", "[PrimeSpeech] RESET received", config.LOG_LEVEL)
                    # Note: Can't actually stop ongoing synthesis, but it's OK
                    # because we only process one segment at a time now
                    send_log(node, "INFO", "[PrimeSpeech] Reset acknowledged", config.LOG_LEVEL)
                
                elif command == "stats":
                    send_log(node, "INFO", f"Total syntheses: {total_syntheses}", config.LOG_LEVEL)
                    send_log(node, "INFO", f"Total audio duration: {total_duration:.1f}s", config.LOG_LEVEL)
        
        elif event["type"] == "STOP":
            break
    
    send_log(node, "INFO", "PrimeSpeech node stopped", config.LOG_LEVEL)


if __name__ == "__main__":
    main()



import time
import logging
import os
import random
import traceback
import numpy as np
import torch
import torch.utils.data
from tqdm import tqdm

from moyoyo_tts.module import commons
from moyoyo_tts.module.mel_processing import spectrogram_torch
from moyoyo_tts.text import cleaned_text_to_sequence
from moyoyo_tts.utils import load_wav_to_torch, load_filepaths_and_text
import torch.nn.functional as F
from functools import lru_cache
import requests
from scipy.io import wavfile
from io import BytesIO
from moyoyo_tts.tools.my_utils import load_audio
version = os.environ.get('version',None)
# ZeroDivisionError fixed by Tybost (https://github.com/RVC-Boss/GPT-SoVITS/issues/79)
class TextAudioSpeakerLoader(torch.utils.data.Dataset):
    """
    1) loads audio, speaker_id, text pairs
    2) normalizes text and converts them to sequences of integers
    3) computes spectrograms from audio files.
    """

    def __init__(self, hparams, val=False):
        exp_dir = hparams.exp_dir
        self.path2 = "%s/2-name2text.txt" % exp_dir
        self.path4 = "%s/4-cnhubert" % exp_dir
        self.path5 = "%s/5-wav32k" % exp_dir
        assert os.path.exists(self.path2)
        assert os.path.exists(self.path4)
        assert os.path.exists(self.path5)
        names4 = set([name[:-3] for name in list(os.listdir(self.path4))])  # 去除.pt后缀
        names5 = set(os.listdir(self.path5))
        self.phoneme_data = {}
        with open(self.path2, "r", encoding="utf8") as f:
            lines = f.read().strip("\n").split("\n")

        for line in lines:
            tmp = line.split("\t")
            if (len(tmp) != 4):
                continue
            self.phoneme_data[tmp[0]] = [tmp[1]]

        self.audiopaths_sid_text = list(set(self.phoneme_data) & names4 & names5)
        tmp = self.audiopaths_sid_text
        leng = len(tmp)
        min_num = 100
        if (leng < min_num):
            self.audiopaths_sid_text = []
            for _ in range(max(2, int(min_num / leng))):
                self.audiopaths_sid_text += tmp
        self.max_wav_value = hparams.max_wav_value
        self.sampling_rate = hparams.sampling_rate
        self.filter_length = hparams.filter_length
        self.hop_length = hparams.hop_length
        self.win_length = hparams.win_length
        self.sampling_rate = hparams.sampling_rate
        self.val = val

        random.seed(1234)
        random.shuffle(self.audiopaths_sid_text)

        print("phoneme_data_len:", len(self.phoneme_data.keys()))
        print("wav_data_len:", len(self.audiopaths_sid_text))

        audiopaths_sid_text_new = []
        lengths = []
        skipped_phone = 0
        skipped_dur = 0
        for audiopath in tqdm(self.audiopaths_sid_text):
            try:
                phoneme = self.phoneme_data[audiopath][0]
                phoneme = phoneme.split(' ')
                phoneme_ids = cleaned_text_to_sequence(phoneme, version)
            except Exception:
                print(f"{audiopath} not in self.phoneme_data !")
                skipped_phone += 1
                continue

            size = os.path.getsize("%s/%s" % (self.path5, audiopath))
            duration = size / self.sampling_rate / 2

            if duration == 0:
                print(f"Zero duration for {audiopath}, skipping...")
                skipped_dur += 1
                continue

            if 54 > duration > 0.6 or self.val:
                audiopaths_sid_text_new.append([audiopath, phoneme_ids])
                lengths.append(size // (2 * self.hop_length))
            else:
                skipped_dur += 1
                continue

        print("skipped_phone: ", skipped_phone, ", skipped_dur: ", skipped_dur)
        print("total left: ", len(audiopaths_sid_text_new))
        assert len(audiopaths_sid_text_new) > 1  # 至少能凑够batch size，这里todo
        self.audiopaths_sid_text = audiopaths_sid_text_new
        self.lengths = lengths

    def get_audio_text_speaker_pair(self, audiopath_sid_text):
        audiopath, phoneme_ids = audiopath_sid_text
        text = torch.FloatTensor(phoneme_ids)
        try:
            spec, wav = self.get_audio("%s/%s" % (self.path5, audiopath))
            with torch.no_grad():
                ssl = torch.load("%s/%s.pt" % (self.path4, audiopath), map_location="cpu")
                if (ssl.shape[-1] != spec.shape[-1]):
                    typee = ssl.dtype
                    ssl = F.pad(ssl.float(), (0, 1), mode="replicate").to(typee)
                ssl.requires_grad = False
        except:
            traceback.print_exc()
            spec = torch.zeros(1025, 100)
            wav = torch.zeros(1, 100 * self.hop_length)
            ssl = torch.zeros(1, 768, 100)
            text = text[-1:]
            print("load audio or ssl error!!!!!!", audiopath)
        return (ssl, spec, wav, text)

    def get_audio(self, filename):
        audio_array = load_audio(filename, self.sampling_rate)  # load_audio的方法是已经归一化到-1~1之间的，不用再/32768
        audio = torch.FloatTensor(audio_array)  # /32768
        audio_norm = audio
        audio_norm = audio_norm.unsqueeze(0)
        spec = spectrogram_torch(audio_norm, self.filter_length, self.sampling_rate, self.hop_length, self.win_length,
                                  center=False)
        spec = torch.squeeze(spec, 0)
        return spec, audio_norm

    def get_sid(self, sid):
        sid = torch.LongTensor([int(sid)])
        return sid

    def __getitem__(self, index):
        # with torch.no_grad():
        return self.get_audio_text_speaker_pair(self.audiopaths_sid_text[index])

    def __len__(self):
        return len(self.audiopaths_sid_text)

    def random_slice(self, ssl, wav, mel):
        assert abs(ssl.shape[-1] - wav.shape[-1] // self.hop_length) < 3, (
        "first", ssl.shape, wav.shape)

        len_mel = mel.shape[1]
        if self.val:
            reference_mel = mel[:, :len_mel // 3]
            return reference_mel, ssl, wav, mel
        dir = random.randint(0, 1)
        sep_point = random.randint(int(len_mel // 3), int(len_mel // 3 * 2))

        if dir == 0:
            reference_mel = mel[:, :sep_point]
            ssl = ssl[:, :, sep_point:]
            wav2 = wav[:, sep_point * self.hop_length:]
            mel = mel[:, sep_point:]
        else:
            reference_mel = mel[:, sep_point:]
            ssl = ssl[:, :, :sep_point]
            wav2 = wav[:, :sep_point * self.hop_length]
            mel = mel[:, :sep_point]

        assert abs(ssl.shape[-1] - wav2.shape[-1] // self.hop_length) < 3, (
        ssl.shape, wav.shape, wav2.shape, mel.shape, sep_point, self.hop_length, sep_point * self.hop_length, dir)
        return reference_mel, ssl, wav2, mel


class TextAudioSpeakerCollate():
    """ Zero-pads model inputs and targets
    """

    def __init__(self, return_ids=False):
        self.return_ids = return_ids

    def __call__(self, batch):
        """Collate's training batch from normalized text, audio and speaker identities
        PARAMS
        ------
        batch: [text_normalized, spec_normalized, wav_normalized, sid]
        """
        # Right zero-pad all one-hot text sequences to max input length
        _, ids_sorted_decreasing = torch.sort(
            torch.LongTensor([x[1].size(1) for x in batch]),
            dim=0, descending=True)

        max_ssl_len = max([x[0].size(2) for x in batch])
        max_ssl_len = int(2 * ((max_ssl_len // 2) + 1))
        max_spec_len = max([x[1].size(1) for x in batch])
        max_spec_len = int(2 * ((max_spec_len // 2) + 1))
        max_wav_len = max([x[2].size(1) for x in batch])
        max_text_len = max([x[3].size(0) for x in batch])

        ssl_lengths = torch.LongTensor(len(batch))
        spec_lengths = torch.LongTensor(len(batch))
        wav_lengths = torch.LongTensor(len(batch))
        text_lengths = torch.LongTensor(len(batch))

        spec_padded = torch.FloatTensor(len(batch), batch[0][1].size(0), max_spec_len)
        wav_padded = torch.FloatTensor(len(batch), 1, max_wav_len)
        ssl_padded = torch.FloatTensor(len(batch), batch[0][0].size(1), max_ssl_len)
        text_padded = torch.LongTensor(len(batch), max_text_len)

        spec_padded.zero_()
        wav_padded.zero_()
        ssl_padded.zero_()
        text_padded.zero_()

        for i in range(len(ids_sorted_decreasing)):
            row = batch[ids_sorted_decreasing[i]]

            ssl = row[0]
            ssl_padded[i, :, :ssl.size(2)] = ssl[0, :, :]
            ssl_lengths[i] = ssl.size(2)

            spec = row[1]
            spec_padded[i, :, :spec.size(1)] = spec
            spec_lengths[i] = spec.size(1)

            wav = row[2]
            wav_padded[i, :, :wav.size(1)] = wav
            wav_lengths[i] = wav.size(1)

            text = row[3]
            text_padded[i, :text.size(0)] = text
            text_lengths[i] = text.size(0)

        return ssl_padded, ssl_lengths, spec_padded, spec_lengths, wav_padded, wav_lengths, text_padded, text_lengths


class DistributedBucketSampler(torch.utils.data.distributed.DistributedSampler):
    """
    Maintain similar input lengths in a batch.
    Length groups are specified by boundaries.
    Ex) boundaries = [b1, b2, b3] -> any batch is included either {x | b1 < length(x) <=b2} or {x | b2 < length(x) <= b3}.

    It removes samples which are not included in the boundaries.
    Ex) boundaries = [b1, b2, b3] -> any x s.t. length(x) <= b1 or length(x) > b3 are discarded.
    """

    def __init__(self, dataset, batch_size, boundaries, num_replicas=None, rank=None, shuffle=True):
        super().__init__(dataset, num_replicas=num_replicas, rank=rank, shuffle=shuffle)
        self.lengths = dataset.lengths
        self.batch_size = batch_size
        self.boundaries = boundaries

        self.buckets, self.num_samples_per_bucket = self._create_buckets()
        self.total_size = sum(self.num_samples_per_bucket)
        self.num_samples = self.total_size // self.num_replicas

    def _create_buckets(self):
        buckets = [[] for _ in range(len(self.boundaries) - 1)]
        for i in range(len(self.lengths)):
            length = self.lengths[i]
            idx_bucket = self._bisect(length)
            if idx_bucket != -1:
                buckets[idx_bucket].append(i)

        i = len(buckets) - 1
        while i >= 0:
            if len(buckets[i]) == 0:
                buckets.pop(i)
                self.boundaries.pop(i + 1)
            i -= 1

        num_samples_per_bucket = []
        for i in range(len(buckets)):
            len_bucket = len(buckets[i])
            total_batch_size = self.num_replicas * self.batch_size
            rem = (total_batch_size - (len_bucket % total_batch_size)) % total_batch_size
            num_samples_per_bucket.append(len_bucket + rem)
        return buckets, num_samples_per_bucket

    def __iter__(self):
        g = torch.Generator()
        g.manual_seed(self.epoch)

        indices = []
        if self.shuffle:
            for bucket in self.buckets:
                indices.append(torch.randperm(len(bucket), generator=g).tolist())
        else:
            for bucket in self.buckets:
                indices.append(list(range(len(bucket))))

        batches = []
        for i in range(len(self.buckets)):
            bucket = self.buckets[i]
            len_bucket = len(bucket)
            ids_bucket = indices[i]
            num_samples_bucket = self.num_samples_per_bucket[i]

            rem = num_samples_bucket - len_bucket
            ids_bucket = ids_bucket + ids_bucket * (rem // len_bucket) + ids_bucket[:(rem % len_bucket)]

            ids_bucket = ids_bucket[self.rank::self.num_replicas]

            for j in range(len(ids_bucket) // self.batch_size):
                batch = [bucket[idx] for idx in ids_bucket[j * self.batch_size:(j + 1) * self.batch_size]]
                batches.append(batch)

        if self.shuffle:
            batch_ids = torch.randperm(len(batches), generator=g).tolist()
            batches = [batches[i] for i in batch_ids]
        self.batches = batches

        assert len(self.batches) * self.batch_size == self.num_samples
        return iter(self.batches)

    def _bisect(self, x, lo=0, hi=None):
        if hi is None:
            hi = len(self.boundaries) - 1

        if hi > lo:
            mid = (hi + lo) // 2
            if self.boundaries[mid] < x and x <= self.boundaries[mid + 1]:
                return mid
            elif x <= self.boundaries[mid]:
                return self._bisect(x, lo, mid)
            else:
                return self._bisect(x, mid + 1, hi)
        else:
            return -1

    def __len__(self):
        return self.num_samples // self.batch_size
